import base64
from enum import Enum

import jwt
import wrapt
from mongoengine.errors import DoesNotExist
from passlib.apps import custom_app_context
from tornado.web import HTTPError

import brew_view
from bg_utils.mongo.models import Principal, Role
from brewtils.errors import RequestForbidden
from brewtils.models import Principal as BrewtilsPrincipal, Role as BrewtilsRole


class Permissions(Enum):
    ALL = "bg-all"
    COMMAND_CREATE = "bg-command-create"
    COMMAND_READ = "bg-command-read"
    COMMAND_UPDATE = "bg-command-update"
    COMMAND_DELETE = "bg-command-delete"
    EVENT_CREATE = "bg-event-create"
    EVENT_READ = "bg-event-read"
    EVENT_UPDATE = "bg-event-update"
    EVENT_DELETE = "bg-event-delete"
    INSTANCE_CREATE = "bg-instance-create"
    INSTANCE_READ = "bg-instance-read"
    INSTANCE_UPDATE = "bg-instance-update"
    INSTANCE_DELETE = "bg-instance-delete"
    QUEUE_CREATE = "bg-queue-create"
    QUEUE_READ = "bg-queue-read"
    QUEUE_UPDATE = "bg-queue-update"
    QUEUE_DELETE = "bg-queue-delete"
    JOB_CREATE = "bg-job-create"
    JOB_READ = "bg-job-read"
    JOB_UPDATE = "bg-job-update"
    JOB_DELETE = "bg-job-delete"
    REQUEST_CREATE = "bg-request-create"
    REQUEST_READ = "bg-request-read"
    REQUEST_UPDATE = "bg-request-update"
    REQUEST_DELETE = "bg-request-delete"
    ROLE_CREATE = "bg-role-create"
    ROLE_READ = "bg-role-read"
    ROLE_UPDATE = "bg-role-update"
    ROLE_DELETE = "bg-role-delete"
    SYSTEM_CREATE = "bg-system-create"
    SYSTEM_READ = "bg-system-read"
    SYSTEM_UPDATE = "bg-system-update"
    SYSTEM_DELETE = "bg-system-delete"
    USER_CREATE = "bg-user-create"
    USER_READ = "bg-user-read"
    USER_UPDATE = "bg-user-update"
    USER_DELETE = "bg-user-delete"


Permissions.values = {p.value for p in Permissions}


def authenticated(permissions=None):
    """Decorator used to require permissions for access to a resource.

    Args:
        permissions: Collection of Permissions enums. Note that if multiple
            permissions are specified then a principal must have all of them
            to access the resource.

    Returns:
        The wrapper function
    """

    @wrapt.decorator
    def wrapper(wrapped, instance, args, kwargs):

        # The interplay between wrapt and gen.coroutine causes things to get
        # a little confused, so we have to be flexible
        handler = instance or args[0]

        check_permission(handler.current_user, permissions)

        return wrapped(*args, **kwargs)

    return wrapper


def check_permission(principal, required_permissions):
    """Determine if a principal has access to a resource

    Args:
        principal: the principal to test
        required_permissions: collection of strings

    Returns:
        None

    Raises:
        HTTPError(status_code=401): The requested resource requires auth
        RequestForbidden(status_code=403): The current principal does not have
            permission to access the requested resource
    """
    if Permissions.ALL.value in principal.permissions:
        return True

    # Convert to strings for easier comparison
    permission_strings = set(p.value for p in required_permissions)

    if not permission_strings.intersection(principal.permissions):
        # Need to make a distinction between "you need to be authenticated
        # to do this" and "you've been authenticated and denied"
        if principal == brew_view.anonymous_principal:
            raise HTTPError(status_code=401)
        else:
            raise RequestForbidden(
                "Action requires permissions %s" % permission_strings
            )


def anonymous_principal():
    """Load correct anonymous permissions

    This exists in a weird space. We need to set the roles attribute to a 'real'
    Role object so it works correctly when the REST handler goes to serialize
    this principal.

    However, we also need to set the permissions attribute to the consolidated
    permission list so that ``check_permission`` will be able to do a comparison
    without having to calculate effective permissions every time.
    """

    if brew_view.config.auth.enabled:
        roles = Principal.objects.get(username="anonymous").roles
    else:
        roles = [Role(name="bg-admin", permissions=["bg-all"])]

    _, permissions = coalesce_permissions(roles)

    return BrewtilsPrincipal(username="anonymous", roles=roles, permissions=permissions)


def coalesce_permissions(role_list):
    """Determine permissions"""

    if not role_list:
        return set(), set()

    aggregate_roles = set()
    aggregate_perms = set()

    for role in role_list:
        aggregate_roles.add(role.name)
        aggregate_perms |= set(role.permissions)

        nested_roles, nested_perms = coalesce_permissions(role.roles)
        aggregate_roles |= nested_roles
        aggregate_perms |= nested_perms

    return aggregate_roles, aggregate_perms


def basic_auth(request):
    """Determine if a basic authorization header is valid

    Args:
        request: The request to authenticate

    Returns:
        Brewtils principal if auth_header is valid, None otherwise
    """
    auth_header = request.headers.get("Authorization")
    if not auth_header or not auth_header.startswith("Basic "):
        return None

    auth_decoded = base64.b64decode(auth_header[6:]).decode()
    username, password = auth_decoded.split(":")

    try:
        principal = Principal.objects.get(username=username)

        if custom_app_context.verify(password, principal.hash):
            return principal
    except DoesNotExist:
        # Don't handle this differently to prevent an attacker from being able
        #  to enumerate a list of user names
        pass

    return None


def bearer_auth(request):
    """Determine a principal from a JWT in the Authorization header

    Args:
        request: The request to authenticate

    Returns:
        Brewtils principal if JWT is valid, None otherwise
    """
    auth_header = request.headers.get("Authorization")
    if not auth_header or not auth_header.startswith("Bearer "):
        return None

    token = auth_header.split(" ")[1]

    return _principal_from_token(token)


def query_token_auth(request):
    """Determine a principal from a JWT in query parameter 'token'

    Args:
        request: The request to authenticate

    Returns:
        Brewtils principal if JWT is valid, None otherwise
    """
    token_args = request.query_arguments.get("token", None)
    if token_args is None:
        return None

    return _principal_from_token(token_args[0])


def _principal_from_token(token):
    """Determine a principal from a JWT

    Args:
        token: The JWT

    Returns:
        Brewtils principal if JWT is valid, None otherwise
    """
    try:
        decoded = jwt.decode(
            token,
            key=brew_view.config.auth.token.secret,
            algorithm=brew_view.config.auth.token.algorithm,
        )
    except jwt.exceptions.ExpiredSignatureError:
        raise HTTPError(status_code=401, log_message="Signature expired")

    return BrewtilsPrincipal(
        id=decoded["sub"],
        username=decoded.get("username", ""),
        roles=[BrewtilsRole(name=role) for role in decoded.get("roles", [])],
        permissions=decoded.get("permissions", []),
    )


class AuthMixin(object):

    auth_providers = [bearer_auth, basic_auth]

    def get_current_user(self):
        """Use registered handlers to determine current user"""

        for provider in self.auth_providers:
            principal = provider(self.request)

            if principal is not None:
                return principal

        return brew_view.anonymous_principal



# -*- coding: utf-8 -*-
"""scheduler module methods."""
from apscheduler.job import Job as APJob
from apscheduler.triggers.cron import CronTrigger
from apscheduler.triggers.date import DateTrigger
from apscheduler.triggers.interval import IntervalTrigger

from pytz import utc


def construct_trigger(trigger_type, bg_trigger):
    """Construct an apscheduler trigger based on type and mongo document."""
    trigger_type = trigger_type
    trigger_kwargs = bg_trigger.get_scheduler_kwargs()
    if trigger_type == "date":
        return DateTrigger(**trigger_kwargs)
    elif trigger_type == "interval":
        return IntervalTrigger(**trigger_kwargs)
    elif trigger_type == "cron":
        return CronTrigger(**trigger_kwargs)
    else:
        raise ValueError("Invalid trigger type %s" % trigger_type)


def db_to_scheduler(document, scheduler, alias="beer_garden"):
    """Convert a database job to a scheduler's job."""
    job = APJob.__new__(APJob)
    if document.next_run_time:
        next_run_time = utc.localize(document.next_run_time)
    else:
        next_run_time = None
    state = {
        "id": document.id,
        "func": "brew_view.scheduler.runner:run_job",
        "trigger": construct_trigger(document.trigger_type, document.trigger),
        "executor": "default",
        "args": (),
        "kwargs": {
            "request_template": document.request_template,
            "job_id": str(document.id),
        },
        "name": document.name,
        "misfire_grace_time": document.misfire_grace_time,
        "coalesce": document.coalesce,
        "max_instances": 3,
        "next_run_time": next_run_time,
    }
    job.__setstate__(state)
    job._scheduler = scheduler
    job._jobstore_alias = alias
    return job



import datetime
from prometheus_client import Gauge, Counter, Summary

from bg_utils.mongo.models import Request
from brewtils.models import Request as BrewtilsRequest

# Summaries:
http_api_latency_total = Summary(
    "bg_http_api_latency_seconds",
    "Total number of seconds each API endpoint is taking to respond.",
    ["method", "route", "status"],
)
plugin_command_latency = Summary(
    "bg_plugin_command_latency_seconds",
    "Total time taken for a command to complete in seconds.",
    ["system", "instance_name", "system_version", "command", "status"],
)

# Counters:
completed_request_counter = Counter(
    "bg_completed_requests_total",
    "Number of completed requests.",
    ["system", "instance_name", "system_version", "command", "status"],
)
request_counter_total = Counter(
    "bg_requests_total",
    "Number of requests.",
    ["system", "instance_name", "system_version", "command"],
)

# Gauges:
queued_request_gauge = Gauge(
    "bg_queued_requests",
    "Number of requests waiting to be processed.",
    ["system", "instance_name", "system_version"],
)
in_progress_request_gauge = Gauge(
    "bg_in_progress_requests",
    "Number of requests IN_PROGRESS",
    ["system", "instance_name", "system_version"],
)


def request_latency(start_time):
    """Measure request latency in seconds as a float."""
    return (datetime.datetime.utcnow() - start_time).total_seconds()


def initialize_counts():
    for request in Request.objects(status__in=["CREATED", "IN_PROGRESS"]):
        label_args = {
            "system": request.system,
            "system_version": request.system_version,
            "instance_name": request.instance_name,
        }

        if request.status == "CREATED":
            queued_request_gauge.labels(**label_args).inc()
        elif request.status == "IN_PROGRESS":
            in_progress_request_gauge.labels(**label_args).inc()


def request_created(request):
    queued_request_gauge.labels(
        system=request.system,
        system_version=request.system_version,
        instance_name=request.instance_name,
    ).inc()
    request_counter_total.labels(
        system=request.system,
        system_version=request.system_version,
        instance_name=request.instance_name,
        command=request.command,
    ).inc()


def request_updated(request, status_before):
    """Update metrics associated with a Request update

    This call should happen after the save to the database.

    """
    if (
        status_before == request.status
        or status_before in BrewtilsRequest.COMPLETED_STATUSES
    ):
        return

    labels = {
        "system": request.system,
        "system_version": request.system_version,
        "instance_name": request.instance_name,
    }

    if status_before == "CREATED":
        queued_request_gauge.labels(**labels).dec()
    elif status_before == "IN_PROGRESS":
        in_progress_request_gauge.labels(**labels).dec()

    if request.status == "IN_PROGRESS":
        in_progress_request_gauge.labels(**labels).inc()

    elif request.status in BrewtilsRequest.COMPLETED_STATUSES:
        latency = request_latency(request.created_at)
        labels["command"] = request.command
        labels["status"] = request.status

        completed_request_counter.labels(**labels).inc()
        plugin_command_latency.labels(**labels).observe(latency)



import logging

import brew_view
from bg_utils.mongo.models import Request, Job
from bg_utils.mongo.parser import MongoParser
from brew_view.authorization import authenticated, Permissions
from brew_view.base_handler import BaseHandler
from brew_view.metrics import request_updated
from brewtils.errors import ModelValidationError
from brewtils.models import Events, Request as BrewtilsRequest


class RequestAPI(BaseHandler):

    parser = MongoParser()
    logger = logging.getLogger(__name__)

    @authenticated(permissions=[Permissions.REQUEST_READ])
    def get(self, request_id):
        """
        ---
        summary: Retrieve a specific Request
        parameters:
          - name: request_id
            in: path
            required: true
            description: The ID of the Request
            type: string
        responses:
          200:
            description: Request with the given ID
            schema:
              $ref: '#/definitions/Request'
          404:
            $ref: '#/definitions/404Error'
          50x:
            $ref: '#/definitions/50xError'
        tags:
          - Requests
        """
        self.logger.debug("Getting Request: %s", request_id)

        req = Request.objects.get(id=str(request_id))
        req.children = Request.objects(parent=req)

        self.write(self.parser.serialize_request(req, to_string=False))

    @authenticated(permissions=[Permissions.REQUEST_UPDATE])
    def patch(self, request_id):
        """
        ---
        summary: Partially update a Request
        description: |
          The body of the request needs to contain a set of instructions detailing the
          updates to apply. Currently the only operation supported is `replace`, with
          paths `/status`, `/output`, and `/error_class`:
          ```JSON
          {
            "operations": [
              { "operation": "replace", "path": "/status", "value": "" }
            ]
          }
          ```
        parameters:
          - name: request_id
            in: path
            required: true
            description: The ID of the Request
            type: string
          - name: patch
            in: body
            required: true
            description: Instructions for how to update the Request
            schema:
              $ref: '#/definitions/Patch'
        responses:
          200:
            description: Request with the given ID
            schema:
              $ref: '#/definitions/Request'
          400:
            $ref: '#/definitions/400Error'
          404:
            $ref: '#/definitions/404Error'
          50x:
            $ref: '#/definitions/50xError'
        tags:
          - Requests
        """
        req = Request.objects.get(id=request_id)
        operations = self.parser.parse_patch(
            self.request.decoded_body, many=True, from_string=True
        )
        wait_event = None

        # We note the status before the operations, because it is possible for the
        # operations to update the status of the request. In that case, because the
        # updates are coming in in a single request it is okay to update the output or
        # error_class. Ideally this would be handled correctly when we better integrate
        # PatchOperations with their models.
        status_before = req.status

        for op in operations:
            if op.operation == "replace":
                if op.path == "/status":
                    if op.value.upper() in BrewtilsRequest.STATUS_LIST:
                        req.status = op.value.upper()

                        if op.value.upper() == "IN_PROGRESS":
                            self.request.event.name = Events.REQUEST_STARTED.name

                        elif op.value.upper() in BrewtilsRequest.COMPLETED_STATUSES:
                            self.request.event.name = Events.REQUEST_COMPLETED.name

                            if request_id in brew_view.request_map:
                                wait_event = brew_view.request_map[request_id]
                    else:
                        error_msg = "Unsupported status value '%s'" % op.value
                        self.logger.warning(error_msg)
                        raise ModelValidationError(error_msg)
                elif op.path == "/output":
                    if req.output == op.value:
                        continue

                    if status_before in Request.COMPLETED_STATUSES:
                        raise ModelValidationError(
                            "Cannot update output for a request "
                            "that is already completed"
                        )
                    req.output = op.value
                elif op.path == "/error_class":
                    if req.error_class == op.value:
                        continue

                    if status_before in Request.COMPLETED_STATUSES:
                        raise ModelValidationError(
                            "Cannot update error_class for a "
                            "request that is already completed"
                        )
                    req.error_class = op.value
                    self.request.event.error = True
                else:
                    error_msg = "Unsupported path '%s'" % op.path
                    self.logger.warning(error_msg)
                    raise ModelValidationError(error_msg)
            else:
                error_msg = "Unsupported operation '%s'" % op.operation
                self.logger.warning(error_msg)
                raise ModelValidationError(error_msg)

        req.save()

        # Metrics
        request_updated(req, status_before)
        self._update_job_numbers(req, status_before)

        if wait_event:
            wait_event.set()

        self.request.event_extras = {"request": req, "patch": operations}

        self.write(self.parser.serialize_request(req, to_string=False))

    def _update_job_numbers(self, request, status_before):
        if (
            not request.metadata.get("_bg_job_id")
            or status_before == request.status
            or request.status not in Request.COMPLETED_STATUSES
        ):
            return

        try:
            job_id = request.metadata.get("_bg_job_id")
            document = Job.objects.get(id=job_id)
            if request.status == "ERROR":
                document.error_count += 1
            elif request.status == "SUCCESS":
                document.success_count += 1
            document.save()
        except Exception as exc:
            self.logger.warning("Could not update job counts.")
            self.logger.exception(exc)



# -*- coding: utf-8 -*-
import logging

from tornado.gen import coroutine

import brew_view
from bg_utils.mongo.models import Job
from bg_utils.mongo.parser import MongoParser
from brew_view.authorization import authenticated, Permissions
from brew_view.base_handler import BaseHandler
from brewtils.errors import ModelValidationError


class JobAPI(BaseHandler):
    logger = logging.getLogger(__name__)
    parser = MongoParser()

    @authenticated(permissions=[Permissions.JOB_READ])
    def get(self, job_id):
        """
        ---
        summary: Retrieve a specific Job
        parameters:
          - name: job_id
            in: path
            required: true
            description: The ID of the Job
            type: string
        responses:
          200:
            description: Job with the given ID
            schema:
              $ref: '#/definitions/Job'
          404:
            $ref: '#/definitions/404Error'
          50x:
            $ref: '#/definitions/50xError'
        tags:
          - Jobs
        """
        document = Job.objects.get(id=job_id)
        self.set_header("Content-Type", "application/json; charset=UTF-8")
        self.write(self.parser.serialize_job(document, to_string=False))

    @authenticated(permissions=[Permissions.JOB_UPDATE])
    def patch(self, job_id):
        """
        ---
        summary: Pause/Resume a job
        description: |
          The body of the request needs to contain a set of instructions
          detailing the actions to take. Currently the only operation
          supported is `update` with `path` of `/status`.


          You can pause a job with:
          ```JSON
          {
            "operations": [
              { "operation": "update", "path": "/status", "value": "PAUSED" }
            ]
          }
          ```

          And resume it with:
          ```JSON
          {
            "operations": [
                { "operation": "update", "path": "/status", "value": "RUNNING" }
            ]
          }
          ```
        parameters:
          - name: job_id
            in: path
            required: true
            description: The ID of the Job
            type: string
          - name: patch
            in: body
            required: true
            description: Instructions for the actions to take
            schema:
              $ref: '#/definitions/Patch'
        responses:
          200:
            description: Job with the given ID
            schema:
              $ref: '#/definitions/Job'
          400:
            $ref: '#/definitions/400Error'
          404:
            $ref: '#/definitions/404Error'
          50x:
            $ref: '#/definitions/50xError'
        tags:
          - Jobs
        """
        job = Job.objects.get(id=job_id)
        operations = self.parser.parse_patch(
            self.request.decoded_body, many=True, from_string=True
        )

        for op in operations:
            if op.operation == "update":
                if op.path == "/status":
                    if str(op.value).upper() == "PAUSED":
                        brew_view.request_scheduler.pause_job(
                            job_id, jobstore="beer_garden"
                        )
                        job.status = "PAUSED"
                    elif str(op.value).upper() == "RUNNING":
                        brew_view.request_scheduler.resume_job(
                            job_id, jobstore="beer_garden"
                        )
                        job.status = "RUNNING"
                    else:
                        error_msg = "Unsupported status value '%s'" % op.value
                        self.logger.warning(error_msg)
                        raise ModelValidationError(error_msg)
                else:
                    error_msg = "Unsupported path value '%s'" % op.path
                    self.logger.warning(error_msg)
                    raise ModelValidationError(error_msg)
            else:
                error_msg = "Unsupported operation '%s'" % op.operation
                self.logger.warning(error_msg)
                raise ModelValidationError(error_msg)

        job.save()
        self.set_header("Content-Type", "application/json; charset=UTF-8")
        self.write(self.parser.serialize_job(job, to_string=False))

    @coroutine
    @authenticated(permissions=[Permissions.JOB_DELETE])
    def delete(self, job_id):
        """
        ---
        summary: Delete a specific Job.
        description: Will remove a specific job. No further executions will occur.
        parameters:
          - name: job_id
            in: path
            required: true
            description: The ID of the Job
            type: string
        responses:
          204:
            description: Job has been successfully deleted.
          404:
            $ref: '#/definitions/404Error'
          50x:
            $ref: '#/definitions/50xError'
        tags:
          - Jobs
        """
        brew_view.request_scheduler.remove_job(job_id, jobstore="beer_garden")
        self.set_status(204)



import logging

from mongoengine.errors import DoesNotExist

import brew_view
from bg_utils.mongo.models import Role
from bg_utils.mongo.parser import MongoParser
from brew_view.authorization import anonymous_principal, authenticated, Permissions
from brew_view.base_handler import BaseHandler
from brewtils.errors import ModelValidationError


class RoleAPI(BaseHandler):

    logger = logging.getLogger(__name__)

    @authenticated(permissions=[Permissions.ROLE_READ])
    def get(self, role_id):
        """
        ---
        summary: Retrieve all specific Role
        parameters:
          - name: role_id
            in: path
            required: true
            description: The ID of the Role
            type: string
        responses:
          200:
            description: Role with the given ID
            schema:
              $ref: '#/definitions/Role'
          404:
            $ref: '#/definitions/404Error'
          50x:
            $ref: '#/definitions/50xError'
        tags:
          - Roles
        """
        self.write(
            MongoParser.serialize_role(
                Role.objects.get(id=str(role_id)), to_string=False
            )
        )

    @authenticated(permissions=[Permissions.ROLE_DELETE])
    def delete(self, role_id):
        """
        ---
        summary: Delete a specific Role
        parameters:
          - name: role_id
            in: path
            required: true
            description: The ID of the Role
            type: string
        responses:
          204:
            description: Role has been successfully deleted
          404:
            $ref: '#/definitions/404Error'
          50x:
            $ref: '#/definitions/50xError'
        tags:
          - Roles
        """
        role = Role.objects.get(id=str(role_id))
        if role.name in ("bg-admin", "bg-anonymous", "bg-plugin"):
            raise ModelValidationError("Unable to remove '%s' role" % role.name)

        role.delete()

        self.set_status(204)

    @authenticated(permissions=[Permissions.ROLE_UPDATE])
    def patch(self, role_id):
        """
        ---
        summary: Partially update a Role
        description: |
          The body of the request needs to contain a set of instructions
          detailing the updates to apply:
          ```JSON
          {
            "operations": [
              { "operation": "add", "path": "/permissions", "value": "ALL" }
            ]
          }
          ```
        parameters:
          - name: role_id
            in: path
            required: true
            description: The ID of the Role
            type: string
          - name: patch
            in: body
            required: true
            description: Instructions for how to update the Role
            schema:
              $ref: '#/definitions/Patch'
        responses:
          200:
            description: Role with the given ID
            schema:
              $ref: '#/definitions/Role'
          400:
            $ref: '#/definitions/400Error'
          404:
            $ref: '#/definitions/404Error'
          50x:
            $ref: '#/definitions/50xError'
        tags:
          - Roles
        """
        role = Role.objects.get(id=str(role_id))
        operations = MongoParser.parse_patch(
            self.request.decoded_body, many=True, from_string=True
        )

        for op in operations:
            if op.path == "/permissions":
                try:
                    if op.operation == "add":
                        role.permissions.append(Permissions(op.value).value)
                    elif op.operation == "remove":
                        role.permissions.remove(Permissions(op.value).value)
                    elif op.operation == "set":
                        role.permissions = [
                            Permissions(perm).value for perm in op.value
                        ]
                    else:
                        raise ModelValidationError(
                            "Unsupported operation '%s'" % op.operation
                        )
                except ValueError:
                    raise ModelValidationError(
                        "Permission '%s' does not exist" % op.value
                    )

            elif op.path == "/roles":
                try:
                    if op.operation == "add":
                        new_nested = Role.objects.get(name=op.value)
                        ensure_no_cycles(role, new_nested)
                        role.roles.append(new_nested)
                    elif op.operation == "remove":
                        role.roles.remove(Role.objects.get(name=op.value))
                    elif op.operation == "set":
                        # Do this one at a time to be super sure about cycles
                        role.roles = []

                        for role_name in op.value:
                            new_role = Role.objects.get(name=role_name)
                            ensure_no_cycles(role, new_role)
                            role.roles.append(new_role)
                    else:
                        raise ModelValidationError(
                            "Unsupported operation '%s'" % op.operation
                        )
                except DoesNotExist:
                    raise ModelValidationError("Role '%s' does not exist" % op.value)

            else:
                raise ModelValidationError("Unsupported path '%s'" % op.path)

        role.save()

        # Any modification to roles will possibly modify the anonymous user
        brew_view.anonymous_principal = anonymous_principal()

        self.write(MongoParser.serialize_role(role, to_string=False))


class RolesAPI(BaseHandler):
    @authenticated(permissions=[Permissions.ROLE_READ])
    def get(self):
        """
        ---
        summary: Retrieve all Roles
        responses:
          200:
            description: All Roles
            schema:
              type: array
              items:
                $ref: '#/definitions/Role'
          50x:
            $ref: '#/definitions/50xError'
        tags:
          - Roles
        """
        self.set_header("Content-Type", "application/json; charset=UTF-8")
        self.write(
            MongoParser.serialize_role(Role.objects.all(), many=True, to_string=True)
        )

    @authenticated(permissions=[Permissions.ROLE_CREATE])
    def post(self):
        """
        ---
        summary: Create a new Role
        parameters:
          - name: role
            in: body
            description: The Role definition
            schema:
              $ref: '#/definitions/Role'
        consumes:
          - application/json
        responses:
          201:
            description: A new Role has been created
            schema:
              $ref: '#/definitions/Role'
          400:
            $ref: '#/definitions/400Error'
          50x:
            $ref: '#/definitions/50xError'
        tags:
          - Roles
        """
        role = MongoParser.parse_role(self.request.decoded_body, from_string=True)

        # Make sure all new permissions are real
        if not set(role.permissions).issubset(Permissions.values):
            invalid = set(role.permissions).difference(Permissions.values)
            raise ModelValidationError("Permissions %s do not exist" % invalid)

        # And the same for nested roles
        nested_roles = []
        for nested_role in role.roles:
            try:
                db_role = Role.objects.get(name=nested_role.name)

                # There shouldn't be any way to construct a cycle with a new
                # role, but check just to be sure
                ensure_no_cycles(role, db_role)

                nested_roles.append(db_role)
            except DoesNotExist:
                raise ModelValidationError(
                    "Role '%s' does not exist" % nested_role.name
                )
        role.roles = nested_roles

        role.save()

        self.set_status(201)
        self.write(MongoParser.serialize_role(role, to_string=False))


def ensure_no_cycles(base_role, new_role):
    """Make sure there are no nested role cycles

    Do this by looking through new_roles's nested roles and making sure
    base_role doesn't appear

    Args:
        base_role: The role that is being modified
        new_role: The new nested role

    Returns:
        None

    Raises:
        ModelValidationError: A cycle was detected
    """
    for role in new_role.roles:
        if role == base_role:
            raise ModelValidationError("Cycle Detected!")

        ensure_no_cycles(base_role, role)



import logging

from tornado.gen import coroutine

import brew_view
from brew_view import thrift_context
from brew_view.base_handler import BaseHandler


class ConfigHandler(BaseHandler):
    def get(self):
        """Subset of configuration options that the frontend needs"""
        configs = {
            "allow_unsafe_templates": brew_view.config.application.allow_unsafe_templates,
            "application_name": brew_view.config.application.name,
            "amq_admin_port": brew_view.config.amq.connections.admin.port,
            "amq_host": brew_view.config.amq.host,
            "amq_port": brew_view.config.amq.connections.message.port,
            "amq_virtual_host": brew_view.config.amq.virtual_host,
            "backend_host": brew_view.config.backend.host,
            "backend_port": brew_view.config.backend.port,
            "icon_default": brew_view.config.application.icon_default,
            "debug_mode": brew_view.config.debug_mode,
            "url_prefix": brew_view.config.web.url_prefix,
            "metrics_url": brew_view.config.metrics.url,
            "auth_enabled": brew_view.config.auth.enabled,
        }
        self.write(configs)


class VersionHandler(BaseHandler):
    @coroutine
    def get(self):
        with thrift_context() as client:
            try:
                bartender_version = yield client.getVersion()
            except Exception as ex:
                logger = logging.getLogger(__name__)
                logger.error("Could not get Bartender Version.")
                logger.exception(ex)
                bartender_version = "unknown"

        self.write(
            {
                "brew_view_version": brew_view.__version__,
                "bartender_version": bartender_version,
                "current_api_version": "v1",
                "supported_api_versions": ["v1"],
            }
        )


class SwaggerConfigHandler(BaseHandler):
    def get(self):
        self.set_header("Content-Type", "application/json; charset=UTF-8")
        self.write(
            {
                "url": brew_view.config.web.url_prefix + "api/v1/spec",
                "validatorUrl": None,
            }
        )


class SpecHandler(BaseHandler):
    def get(self):
        self.set_header("Content-Type", "application/json; charset=UTF-8")
        self.write(brew_view.api_spec.to_dict())



import logging
from datetime import datetime

from tornado.gen import coroutine

from bg_utils.mongo.models import Instance
from bg_utils.mongo.parser import MongoParser
from brew_view import thrift_context
from brew_view.authorization import authenticated, Permissions
from brew_view.base_handler import BaseHandler
from brewtils.errors import ModelValidationError
from brewtils.models import Events


class InstanceAPI(BaseHandler):

    parser = MongoParser()
    logger = logging.getLogger(__name__)

    event_dict = {
        "initialize": Events.INSTANCE_INITIALIZED.name,
        "start": Events.INSTANCE_STARTED.name,
        "stop": Events.INSTANCE_STOPPED.name,
    }

    @authenticated(permissions=[Permissions.INSTANCE_READ])
    def get(self, instance_id):
        """
        ---
        summary: Retrieve a specific Instance
        parameters:
          - name: instance_id
            in: path
            required: true
            description: The ID of the Instance
            type: string
        responses:
          200:
            description: Instance with the given ID
            schema:
              $ref: '#/definitions/Instance'
          404:
            $ref: '#/definitions/404Error'
          50x:
            $ref: '#/definitions/50xError'
        tags:
          - Instances
        """
        self.logger.debug("Getting Instance: %s", instance_id)

        self.write(
            self.parser.serialize_instance(
                Instance.objects.get(id=instance_id), to_string=False
            )
        )

    @authenticated(permissions=[Permissions.INSTANCE_DELETE])
    def delete(self, instance_id):
        """
        ---
        summary: Delete a specific Instance
        parameters:
          - name: instance_id
            in: path
            required: true
            description: The ID of the Instance
            type: string
        responses:
          204:
            description: Instance has been successfully deleted
          404:
            $ref: '#/definitions/404Error'
          50x:
            $ref: '#/definitions/50xError'
        tags:
          - Instances
        """
        self.logger.debug("Deleting Instance: %s", instance_id)

        Instance.objects.get(id=instance_id).delete()

        self.set_status(204)

    @coroutine
    @authenticated(permissions=[Permissions.INSTANCE_UPDATE])
    def patch(self, instance_id):
        """
        ---
        summary: Partially update an Instance
        description: |
          The body of the request needs to contain a set of instructions detailing the
          updates to apply. Currently the only operations are:

          * initialize
          * start
          * stop
          * heartbeat

          ```JSON
          {
            "operations": [
              { "operation": "" }
            ]
          }
          ```
        parameters:
          - name: instance_id
            in: path
            required: true
            description: The ID of the Instance
            type: string
          - name: patch
            in: body
            required: true
            description: Instructions for how to update the Instance
            schema:
              $ref: '#/definitions/Patch'
        responses:
          200:
            description: Instance with the given ID
            schema:
              $ref: '#/definitions/Instance'
          400:
            $ref: '#/definitions/400Error'
          404:
            $ref: '#/definitions/404Error'
          50x:
            $ref: '#/definitions/50xError'
        tags:
          - Instances
        """
        response = {}
        instance = Instance.objects.get(id=instance_id)
        operations = self.parser.parse_patch(
            self.request.decoded_body, many=True, from_string=True
        )

        for op in operations:
            if op.operation.lower() in ("initialize", "start", "stop"):
                self.request.event.name = self.event_dict[op.operation.lower()]
                with thrift_context() as client:
                    response = yield getattr(client, op.operation.lower() + "Instance")(
                        instance_id
                    )

            elif op.operation.lower() == "heartbeat":
                instance.status_info.heartbeat = datetime.utcnow()
                instance.save()
                response = self.parser.serialize_instance(instance, to_string=False)

            elif op.operation.lower() == "replace":
                if op.path.lower() == "/status":
                    if op.value.upper() == "INITIALIZING":
                        self.request.event.name = Events.INSTANCE_INITIALIZED.name
                        with thrift_context() as client:
                            response = yield client.initializeInstance(instance_id)

                    elif op.value.upper() == "STOPPING":
                        self.request.event.name = Events.INSTANCE_STOPPED.name
                        with thrift_context() as client:
                            response = yield client.stopInstance(instance_id)

                    elif op.value.upper() == "STARTING":
                        self.request.event.name = Events.INSTANCE_STARTED.name
                        with thrift_context() as client:
                            response = yield client.startInstance(instance_id)

                    elif op.value.upper() in ["RUNNING", "STOPPED"]:
                        instance.status = op.value.upper()
                        instance.save()
                        response = self.parser.serialize_instance(
                            instance, to_string=False
                        )

                    else:
                        error_msg = "Unsupported status value '%s'" % op.value
                        self.logger.warning(error_msg)
                        raise ModelValidationError("value", error_msg)
                else:
                    error_msg = "Unsupported path '%s'" % op.path
                    self.logger.warning(error_msg)
                    raise ModelValidationError("value", error_msg)
            else:
                error_msg = "Unsupported operation '%s'" % op.operation
                self.logger.warning(error_msg)
                raise ModelValidationError("value", error_msg)

        if self.request.event.name:
            self.request.event_extras = {"instance": instance}

        self.write(response)



# -*- coding: utf-8 -*-

from apscheduler.jobstores.base import BaseJobStore
from mongoengine import DoesNotExist
from pytz import utc

from bg_utils.mongo.models import Job as BGJob
from brew_view.scheduler import db_to_scheduler


class BGJobStore(BaseJobStore):
    def lookup_job(self, job_id):
        """Get job from mongo, convert it to an apscheduler Job.

        Args:
            job_id: The ID of the job to get.

        Returns:
            An apscheduler job or None.
        """
        try:
            document = BGJob.objects.get(id=job_id)
            return db_to_scheduler(document, self._scheduler, self._alias)
        except DoesNotExist:
            return None

    def get_due_jobs(self, now):
        """Find due jobs and convert them to apscheduler jobs."""
        return self._get_jobs({"next_run_time__lte": now})

    def get_next_run_time(self):
        """Get the next run time as a localized datetime."""
        try:
            document = (
                BGJob.objects(next_run_time__ne=None)
                .order_by("next_run_time")
                .fields(next_run_time=1)
                .first()
            )

            if document:
                return utc.localize(document.next_run_time)
        except DoesNotExist:
            return None

    def get_all_jobs(self):
        """Get all jobs in apscheduler speak."""
        return self._get_jobs({})

    def add_job(self, job):
        """Just updates the next_run_time.

        Notes:

            The jobstore only needs to update the object that has already
            been saved to the database. It is slightly tricky to generate
            the ``next_run_time`` on the apscheduler job, so we just let
            the scheduler do it for us. After that, we update our job's
            next_run_time to be whatever the scheduler set.

        Args:
            job: The job from the scheduler
        """
        document = BGJob.objects.get(id=job.id)
        document.next_run_time = job.next_run_time
        document.save()

    def update_job(self, job):
        """Update the next_run_time for the job."""
        document = BGJob.objects.get(id=job.id)
        document.next_run_time = job.next_run_time
        document.save()

    def remove_job(self, job_id):
        """Remove job with the given ID."""
        BGJob.objects.get(id=job_id).delete()

    def remove_all_jobs(self):
        """Remove all jobs."""
        BGJob.objects.delete()

    def _get_jobs(self, conditions):
        jobs = []
        failed_jobs = []
        for document in BGJob.objects(**conditions).order_by("next_run_time"):
            try:
                jobs.append(db_to_scheduler(document, self._scheduler, self._alias))
            except BaseException:
                self._logger.exception(
                    'Unable to restore job "%s" -- removing it' % document.id
                )
                failed_jobs.append(document)

        # Remove all the jobs we failed to restore
        if failed_jobs:
            for document in failed_jobs:
                document.delete()

        return jobs



SPECIFICATION = {
    "configuration": {
        "type": "dict",
        "bootstrap": True,
        "items": {
            "file": {
                "type": "str",
                "description": "Path to configuration file to use",
                "required": False,
                "cli_short_name": "c",
                "bootstrap": True,
                "previous_names": ["config"],
                "alt_env_names": ["CONFIG"],
            },
            "type": {
                "type": "str",
                "description": "Configuration file type",
                "required": False,
                "cli_short_name": "t",
                "bootstrap": True,
                "choices": ["json", "yaml"],
            },
        },
    },
    "amq": {
        "type": "dict",
        "items": {
            "host": {
                "type": "str",
                "default": "localhost",
                "description": "Hostname of AMQ to use",
                "previous_names": ["amq_host"],
            },
            "heartbeat_interval": {
                "type": "int",
                "default": 3600,
                "description": "Heartbeat interval for AMQ",
                "previous_names": ["amq_heartbeat_interval"],
            },
            "connection_attempts": {
                "type": "int",
                "default": 3,
                "description": "Number of retries to connect to AMQ",
                "previous_names": ["amq_connection_attempts"],
            },
            "exchange": {
                "type": "str",
                "default": "beer_garden",
                "description": "Exchange name to use for AMQ",
                "previous_names": ["amq_exchange"],
            },
            "virtual_host": {
                "type": "str",
                "default": "/",
                "description": "Virtual host to use for AMQ",
                "previous_names": ["amq_virtual_host"],
            },
            "connections": {
                "type": "dict",
                "items": {
                    "admin": {
                        "type": "dict",
                        "items": {
                            "port": {
                                "type": "int",
                                "default": 15672,
                                "description": "Port of the AMQ Admin host",
                                "previous_names": ["amq_admin_port"],
                                "alt_env_names": ["AMQ_ADMIN_PORT"],
                            },
                            "user": {
                                "type": "str",
                                "default": "guest",
                                "description": "Username to login to the AMQ admin",
                                "previous_names": ["amq_admin_user"],
                                "alt_env_names": ["AMQ_ADMIN_USER"],
                            },
                            "password": {
                                "type": "str",
                                "default": "guest",
                                "description": "Password to login to the AMQ admin",
                                "previous_names": [
                                    "amq_admin_password",
                                    "amq_admin_pw",
                                ],
                                "alt_env_names": ["AMQ_ADMIN_PASSWORD", "AMQ_ADMIN_PW"],
                            },
                        },
                    },
                    "message": {
                        "type": "dict",
                        "items": {
                            "port": {
                                "type": "int",
                                "default": 5672,
                                "description": "Port of the AMQ host",
                                "previous_names": ["amq_port"],
                                "alt_env_names": ["AMQ_PORT"],
                            },
                            "password": {
                                "type": "str",
                                "default": "guest",
                                "description": "Password to login to the AMQ host",
                                "previous_names": ["amq_password"],
                                "alt_env_names": ["AMQ_PASSWORD"],
                            },
                            "user": {
                                "type": "str",
                                "default": "guest",
                                "description": "Username to login to the AMQ host",
                                "previous_names": ["amq_user"],
                                "alt_env_names": ["AMQ_USER"],
                            },
                            "ssl": {
                                "type": "dict",
                                "items": {
                                    "enabled": {
                                        "type": "bool",
                                        "default": False,
                                        "description": "Should the connection use SSL",
                                    },
                                    "ca_cert": {
                                        "type": "str",
                                        "description": (
                                            "Path to CA certificate file to use"
                                        ),
                                        "required": False,
                                    },
                                    "ca_verify": {
                                        "type": "bool",
                                        "default": True,
                                        "description": "Verify external certificates",
                                        "required": False,
                                    },
                                    "client_cert": {
                                        "type": "str",
                                        "description": (
                                            "Path to client combined key / certificate"
                                        ),
                                        "required": False,
                                    },
                                },
                            },
                        },
                    },
                },
            },
        },
    },
    "application": {
        "type": "dict",
        "items": {
            "name": {
                "type": "str",
                "default": "Beer Garden",
                "description": "The title to display on the GUI",
                "previous_names": ["application_name"],
            },
            "icon_default": {
                "type": "str",
                "description": "Default font-awesome icon to display",
                "default": "fa-beer",
                "previous_names": ["icon_default"],
                "alt_env_names": ["ICON_DEFAULT"],
            },
            "allow_unsafe_templates": {
                "type": "bool",
                "default": False,
                "description": "Allow unsafe templates to be loaded by the application",
                "previous_names": [
                    "ALLOW_UNSANITIZED_TEMPLATES",
                    "allow_unsafe_templates",
                ],
                "alt_env_names": [
                    "ALLOW_UNSANITIZED_TEMPLATES",
                    "BG_ALLOW_UNSAFE_TEMPLATES",
                ],
            },
        },
    },
    "auth": {
        "type": "dict",
        "items": {
            "enabled": {
                "type": "bool",
                "default": False,
                "description": "Use role-based authentication / authorization",
            },
            "token": {
                "type": "dict",
                "items": {
                    "algorithm": {
                        "type": "str",
                        "default": "HS256",
                        "description": "Algorithm to use when signing tokens",
                    },
                    "lifetime": {
                        "type": "int",
                        "default": 1200,
                        "description": "Time (seconds) before a token expires",
                    },
                    "secret": {
                        "type": "str",
                        "required": False,
                        "description": "Secret to use when signing tokens",
                        "default": "",
                    },
                },
            },
        },
    },
    "backend": {
        "type": "dict",
        "items": {
            "host": {
                "type": "str",
                "default": "localhost",
                "description": "Backend (Bartender) hostname",
                "previous_names": ["backend_host"],
            },
            "port": {
                "type": "int",
                "default": 9090,
                "description": "Backend (Bartender) thrift port",
                "previous_names": ["backend_port"],
            },
            "socket_timeout": {
                "type": "int",
                "default": 13000,
                "description": "Time (in ms) to wait for backend to respond",
                "previous_names": ["backend_socket_timeout"],
                "previous_defaults": [3000],
            },
        },
    },
    "db": {
        "type": "dict",
        "items": {
            "name": {
                "type": "str",
                "default": "beer_garden",
                "description": "Name of the database to use",
                "previous_names": ["db_name"],
            },
            "connection": {
                "type": "dict",
                "items": {
                    "host": {
                        "type": "str",
                        "default": "localhost",
                        "description": "Hostname/IP of the database server",
                        "previous_names": ["db_host"],
                        "alt_env_names": ["DB_HOST"],
                    },
                    "password": {
                        "type": "str",
                        "default": None,
                        "required": False,
                        "description": "Password to connect to the database",
                        "previous_names": ["db_password"],
                        "alt_env_names": ["DB_PASSWORD"],
                    },
                    "port": {
                        "type": "int",
                        "default": 27017,
                        "description": "Port of the database server",
                        "previous_names": ["db_port"],
                        "alt_env_names": ["DB_PORT"],
                    },
                    "username": {
                        "type": "str",
                        "default": None,
                        "required": False,
                        "description": "Username to connect to the database",
                        "previous_names": ["db_username"],
                        "alt_env_names": ["DB_USERNAME"],
                    },
                },
            },
        },
    },
    "cors_enabled": {
        "type": "bool",
        "default": False,
        "description": "Determine if CORS should be enabled",
        "previous_names": ["CORS_ENABLED"],
    },
    "debug_mode": {
        "type": "bool",
        "default": False,
        "description": "Run the application in debug mode",
    },
    "event": {
        "type": "dict",
        "items": {
            "amq": {
                "type": "dict",
                "items": {
                    "enable": {
                        "type": "bool",
                        "default": False,
                        "description": "Publish events to RabbitMQ",
                    },
                    "exchange": {
                        "type": "str",
                        "required": False,
                        "description": "Exchange to use for AMQ events",
                        "previous_names": ["event_amq_exchange"],
                    },
                    "virtual_host": {
                        "type": "str",
                        "default": "/",
                        "required": False,
                        "description": "Virtual host to use for AMQ events",
                        "previous_names": ["event_amq_virtual_host"],
                    },
                },
            },
            "mongo": {
                "type": "dict",
                "items": {
                    "enable": {
                        "type": "bool",
                        "default": True,
                        "description": "Persist events to Mongo",
                        "previous_names": ["event_persist_mongo"],
                        "alt_env_names": ["EVENT_PERSIST_MONGO"],
                    }
                },
            },
        },
    },
    "log": {
        "type": "dict",
        "items": {
            "config_file": {
                "type": "str",
                "description": "Path to a logging configuration file",
                "required": False,
                "cli_short_name": "l",
                "previous_names": ["log_config"],
                "alt_env_names": ["LOG_CONFIG"],
            },
            "file": {
                "type": "str",
                "description": "File you would like the application to log to",
                "required": False,
                "previous_names": ["log_file"],
            },
            "level": {
                "type": "str",
                "description": "Log level for the application",
                "default": "INFO",
                "choices": ["DEBUG", "INFO", "WARN", "WARNING", "ERROR", "CRITICAL"],
                "previous_names": ["log_level"],
            },
        },
    },
    "metrics": {
        "type": "dict",
        "items": {
            "port": {
                "type": "int",
                "description": "Port for prometheus server to listen on.",
                "default": 2338,
            },
            "url": {
                "type": "str",
                "description": "URL to prometheus/grafana server.",
                "required": False,
            },
        },
    },
    "plugin_logging": {
        "type": "dict",
        "items": {
            "config_file": {
                "type": "str",
                "description": "Path to a logging configuration file for plugins",
                "required": False,
                "previous_names": ["plugin_log_config"],
                "alt_env_names": ["PLUGIN_LOG_CONFIG"],
            },
            "level": {
                "type": "str",
                "description": "Default log level for plugins (could be "
                "overwritten by plugin_log_config value)",
                "default": "INFO",
                "choices": ["DEBUG", "INFO", "WARN", "WARNING", "ERROR", "CRITICAL"],
                "previous_names": ["plugin_log_level"],
                "alt_env_names": ["PLUGIN_LOG_LEVEL"],
            },
        },
    },
    "scheduler": {
        "type": "dict",
        "items": {
            "auth": {
                "type": "dict",
                "items": {
                    "username": {
                        "type": "str",
                        "description": "Username that scheduler will use for "
                        "authentication (needs bg-admin role)",
                        "required": False,
                    },
                    "password": {
                        "type": "str",
                        "description": "Password that scheduler will use for "
                        "authentication (needs bg-admin role)",
                        "required": False,
                    },
                },
            },
            "max_workers": {
                "type": "int",
                "default": 10,
                "description": "Number of workers (processes) to run concurrently.",
            },
            "job_defaults": {
                "type": "dict",
                "items": {
                    "coalesce": {
                        "type": "bool",
                        "default": True,
                        "description": (
                            "Should jobs run only once if multiple have missed their "
                            "window"
                        ),
                    },
                    "max_instances": {
                        "type": "int",
                        "default": 3,
                        "description": (
                            "Default maximum instances of a job to run concurrently."
                        ),
                    },
                },
            },
        },
    },
    "web": {
        "type": "dict",
        "items": {
            "ssl": {
                "type": "dict",
                "items": {
                    "enabled": {
                        "type": "bool",
                        "default": False,
                        "description": "Serve content using SSL",
                        "previous_names": ["ssl_enabled"],
                        "alt_env_names": ["SSL_ENABLED"],
                        "cli_separator": "_",
                    },
                    "private_key": {
                        "type": "str",
                        "description": "Path to a private key",
                        "required": False,
                        "previous_names": ["ssl_private_key"],
                        "alt_env_names": ["SSL_PRIVATE_KEY"],
                    },
                    "public_key": {
                        "type": "str",
                        "description": "Path to a public key",
                        "required": False,
                        "previous_names": ["ssl_public_key"],
                        "alt_env_names": ["SSL_PUBLIC_KEY"],
                    },
                    "ca_cert": {
                        "type": "str",
                        "description": (
                            "Path to CA certificate file to use for SSLContext"
                        ),
                        "required": False,
                        "previous_names": ["ca_cert"],
                        "alt_env_names": ["CA_CERT"],
                    },
                    "ca_path": {
                        "type": "str",
                        "description": (
                            "Path to CA certificate path to use for SSLContext"
                        ),
                        "required": False,
                        "previous_names": ["ca_path"],
                        "alt_env_names": ["CA_PATH"],
                    },
                    "client_cert_verify": {
                        "type": "str",
                        "description": (
                            "Client certificate mode to use when handling requests"
                        ),
                        "choices": ["NONE", "OPTIONAL", "REQUIRED"],
                        "default": "NONE",
                        "previous_names": ["client_cert_verify"],
                        "alt_env_names": ["CLIENT_CERT_VERIFY"],
                    },
                },
            },
            "port": {
                "type": "int",
                "default": 2337,
                "description": "Serve content on this port",
                "previous_names": ["web_port"],
            },
            "url_prefix": {
                "type": "str",
                "default": None,
                "description": "URL path prefix",
                "required": False,
                "previous_names": ["url_prefix"],
                "alt_env_names": ["URL_PREFIX"],
            },
            "host": {
                "type": "str",
                "default": "0.0.0.0",
                "description": "Host for the HTTP Server to bind to",
            },
            "public_fqdn": {
                "type": "str",
                "default": "localhost",
                "description": "Public fully-qualified domain name",
                "previous_names": ["public_fqdn"],
                "alt_env_names": ["PUBLIC_FQDN"],
            },
        },
    },
    "shutdown_timeout": {
        "type": "int",
        "default": 5,
        "description": "How long to wait for Brew View to shutdown before terminating",
    },
}


def get_default_logging_config(level, filename):
    if filename:
        brew_view_handler = {
            "class": "logging.handlers.RotatingFileHandler",
            "level": level,
            "formatter": "simple",
            "filename": filename,
            "maxBytes": 10485760,
            "backupCount": 20,
            "encoding": "utf8",
        }
    else:
        brew_view_handler = {
            "class": "logging.StreamHandler",
            "formatter": "simple",
            "level": level,
            "stream": "ext://sys.stdout",
        }

    return {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "simple": {"format": "%(asctime)s - %(name)s - %(levelname)s - %(message)s"}
        },
        "handlers": {"brew-view": brew_view_handler},
        "loggers": {
            "tornado.access": {"level": "WARN"},
            "tornado.application": {"level": "WARN"},
            "tornado.general": {"level": "WARN"},
        },
        "root": {"level": level, "handlers": ["brew-view"]},
    }



import logging

from bg_utils.mongo.models import Command
from bg_utils.mongo.parser import MongoParser
from brew_view.authorization import authenticated, Permissions
from brew_view.base_handler import BaseHandler


class CommandAPI(BaseHandler):

    parser = MongoParser()
    logger = logging.getLogger(__name__)

    @authenticated(permissions=[Permissions.COMMAND_READ])
    def get(self, command_id):
        """
        ---
        summary: Retrieve a specific Command
        parameters:
          - name: command_id
            in: path
            required: true
            description: The ID of the Command
            type: string
        responses:
          200:
            description: Command with the given ID
            schema:
              $ref: '#/definitions/Command'
          404:
            $ref: '#/definitions/404Error'
          50x:
            $ref: '#/definitions/50xError'
        tags:
          - Commands
        """
        self.logger.debug("Getting Command: %s", command_id)

        self.write(
            self.parser.serialize_command(
                Command.objects.get(id=str(command_id)), to_string=False
            )
        )



import json
import logging
from concurrent.futures import ProcessPoolExecutor
from datetime import datetime, timedelta

import jwt
from mongoengine.errors import DoesNotExist
from passlib.apps import custom_app_context
from tornado.gen import coroutine
from tornado.web import HTTPError

import brew_view
from bg_utils.mongo.models import Principal, RefreshToken
from brew_view.authorization import coalesce_permissions
from brew_view.base_handler import BaseHandler


def verify(password, password_hash):
    return custom_app_context.verify(password, password_hash)


class TokenAPI(BaseHandler):

    logger = logging.getLogger(__name__)

    def get(self, token_id):
        """
        ---
        summary: Use a refresh token to retrieve a new access token
        parameters:
          - name: token_id
            in: path
            required: true
            description: The ID of the Token
            type: string
        responses:
          200:
            description: System with the given ID
            schema:
              $ref: '#/definitions/System'
          404:
            $ref: '#/definitions/404Error'
          50x:
            $ref: '#/definitions/50xError'
        tags:
          - Tokens
        """
        try:
            refresh = RefreshToken.objects.get(id=token_id)

            now = datetime.utcnow()
            if now < refresh.expires:
                self.write(
                    json.dumps({"token": generate_access_token(refresh.payload)})
                )
                return
        except DoesNotExist:
            pass

        raise HTTPError(status_code=403, log_message="Bad credentials")

    def delete(self, token_id):
        """
        ---
        summary: Remove a refresh token
        parameters:
          - name: token_id
            in: path
            required: true
            description: The ID of the Token
            type: string
        responses:
          204:
            description: Token has been successfully deleted
          404:
            $ref: '#/definitions/404Error'
          50x:
            $ref: '#/definitions/50xError'
        tags:
          - Tokens
        """
        RefreshToken.objects.get(id=token_id).delete()

        self.set_status(204)


class TokenListAPI(BaseHandler):

    logger = logging.getLogger(__name__)

    def __init__(self, *args, **kwargs):
        super(TokenListAPI, self).__init__(*args, **kwargs)

        self.executor = ProcessPoolExecutor()

    @coroutine
    def post(self):
        """
        ---
        summary: Use credentials to generate access and refresh tokens
        responses:
          200:
            description: All Tokens
            schema:
              type: array
              items:
                $ref: '#/definitions/Command'
          50x:
            $ref: '#/definitions/50xError'
        tags:
          - Tokens
        """
        parsed_body = json.loads(self.request.decoded_body)

        try:
            principal = Principal.objects.get(username=parsed_body["username"])

            verified = yield self.executor.submit(
                verify, str(parsed_body["password"]), str(principal.hash)
            )

            if verified:
                self.write(json.dumps(generate_tokens(principal)))
                return
        except DoesNotExist:
            # Still attempt to verify something so the request takes a while
            custom_app_context.verify("", None)

        raise HTTPError(status_code=403, log_message="Bad credentials")


def generate_tokens(principal):

    roles, permissions = coalesce_permissions(principal.roles)

    payload = {
        "sub": str(principal.id),
        "username": principal.username,
        "roles": list(roles),
        "permissions": list(permissions),
    }

    return {
        "token": generate_access_token(payload),
        "refresh": generate_refresh_token(payload),
    }


def generate_access_token(payload, issue_time=None):
    issue_time = issue_time or datetime.utcnow()

    access_payload = payload.copy()
    access_payload.update(
        {
            "iat": issue_time,
            "exp": issue_time + timedelta(seconds=brew_view.config.auth.token.lifetime),
        }
    )

    return jwt.encode(
        access_payload,
        key=brew_view.config.auth.token.secret,
        algorithm=brew_view.config.auth.token.algorithm,
    ).decode()


def generate_refresh_token(payload, issue_time=None):

    issue_time = issue_time or datetime.utcnow()

    token = RefreshToken(
        issued=issue_time, expires=issue_time + timedelta(hours=24), payload=payload
    )
    token.save()

    return str(token.id)



import logging

from tornado.gen import coroutine

from bg_utils.mongo.models import System
from bg_utils.mongo.parser import MongoParser
from brew_view import thrift_context
from brew_view.authorization import authenticated, Permissions
from brew_view.base_handler import BaseHandler
from brewtils.models import Events, Queue


class QueueListAPI(BaseHandler):

    parser = MongoParser()
    logger = logging.getLogger(__name__)

    @coroutine
    @authenticated(permissions=[Permissions.QUEUE_READ])
    def get(self):
        """
        ---
        summary: Retrieve all queue information
        responses:
          200:
            description: List of all queue information objects
            schema:
              type: array
              items:
                $ref: '#/definitions/Queue'
          50x:
            $ref: '#/definitions/50xError'
        tags:
          - Queues
        """
        self.logger.debug("Getting all queues")

        queues = []
        systems = System.objects.all().select_related(max_depth=1)

        for system in systems:
            for instance in system.instances:

                queue = Queue(
                    name="UNKNOWN",
                    system=system.name,
                    version=system.version,
                    instance=instance.name,
                    system_id=str(system.id),
                    display=system.display_name,
                    size=-1,
                )

                with thrift_context() as client:
                    try:
                        queue_info = yield client.getQueueInfo(
                            system.name, system.version, instance.name
                        )
                        queue.name = queue_info.name
                        queue.size = queue_info.size
                    except Exception:
                        self.logger.error(
                            "Error getting queue size for %s[%s]-%s"
                            % (system.name, instance.name, system.version)
                        )

                queues.append(queue)

        self.set_header("Content-Type", "application/json; charset=UTF-8")
        self.write(self.parser.serialize_queue(queues, to_string=True, many=True))

    @coroutine
    @authenticated(permissions=[Permissions.QUEUE_DELETE])
    def delete(self):
        """
        ---
        summary: Cancel and clear all requests in all queues
        responses:
          204:
            description: All queues successfully cleared
          50x:
            $ref: '#/definitions/50xError'
        tags:
          - Queues
        """
        self.request.event.name = Events.ALL_QUEUES_CLEARED.name

        with thrift_context() as client:
            yield client.clearAllQueues()

        self.set_status(204)


class OldQueueListAPI(BaseHandler):

    logger = logging.getLogger(__name__)

    def get(self):
        """
        ---
        summary: Retrieve all queue information
        deprecated: true
        description: This endpoint is DEPRECATED - Use /api/v1/queues instead.
        responses:
          200:
            description: List of all queue information objects
            schema:
              type: array
              items:
                $ref: '#/definitions/Queue'
          50x:
            $ref: '#/definitions/50xError'
        tags:
          - Deprecated
        """
        self.redirect("/api/v1/queues/", permanent=True)

    def delete(self):
        """
        ---
        summary: Cancel and clear all requests in all queues
        deprecated: true
        description: This endpoint is DEPRECATED - Use /api/v1/queues instead.
        responses:
          204:
            description: All queues successfully cleared
          50x:
            $ref: '#/definitions/50xError'
        tags:
          - Deprecated
        """
        self.redirect("/api/v1/queues/", permanent=True)



# flake8: noqa
"""Module to make imports of Controllers easier"""
from brew_view.controllers.admin_api import AdminAPI, OldAdminAPI
from brew_view.controllers.command_api import CommandAPI
from brew_view.controllers.command_list_api import CommandListAPI
from brew_view.controllers.event_api import EventPublisherAPI, EventSocket
from brew_view.controllers.instance_api import InstanceAPI
from brew_view.controllers.job_api import JobAPI
from brew_view.controllers.job_list_api import JobListAPI
from brew_view.controllers.logging_api import LoggingConfigAPI
from brew_view.controllers.misc_controllers import (
    ConfigHandler,
    VersionHandler,
    SpecHandler,
    SwaggerConfigHandler,
)
from brew_view.controllers.permissions_api import PermissionsAPI
from brew_view.controllers.queue_api import QueueAPI, OldQueueAPI
from brew_view.controllers.queue_list_api import QueueListAPI, OldQueueListAPI
from brew_view.controllers.request_api import RequestAPI
from brew_view.controllers.request_list_api import RequestListAPI
from brew_view.controllers.roles_api import RoleAPI, RolesAPI
from brew_view.controllers.system_api import SystemAPI
from brew_view.controllers.system_list_api import SystemListAPI
from brew_view.controllers.token_api import TokenAPI, TokenListAPI
from brew_view.controllers.users_api import UserAPI, UsersAPI



import logging

from tornado.gen import coroutine
from tornado.locks import Lock

from bg_utils.mongo.models import System, Instance
from bg_utils.mongo.parser import MongoParser
from brew_view.authorization import authenticated, Permissions
from brew_view.base_handler import BaseHandler
from brewtils.errors import ModelValidationError
from brewtils.models import Events
from brewtils.schemas import SystemSchema


class SystemListAPI(BaseHandler):

    parser = MongoParser()
    logger = logging.getLogger(__name__)

    REQUEST_FIELDS = set(SystemSchema.get_attribute_names())

    # Need to ensure that Systems are updated atomically
    system_lock = Lock()

    @authenticated(permissions=[Permissions.SYSTEM_READ])
    def get(self):
        """
        ---
        summary: Retrieve all Systems
        description: |
          This endpoint allows for querying Systems.

          There are several parameters that control which fields are returned
          and what information is available. Things to be aware of:

          * The `include_commands` parameter is __deprecated__. Don't use it.
            Use `exclude_fields=commands` instead.

          * It's possible to specify `include_fields` _and_ `exclude_fields`.
            This doesn't make a lot of sense, but you can do it. If the same
            field is in both `exclude_fields` takes priority (the field will
            NOT be included in the response).

          Systems matching specific criteria can be filtered using additional
          query parameters. This is a very basic capability:

          * ?name=foo&version=1.0.0
            This will return the system named 'foo' with version '1.0.0'
          * ?name=foo&name=bar
            This will not do what you expect: only return the system named
            'bar' will be returned.
        parameters:
          - name: include_fields
            in: query
            required: false
            description: Specify fields to include in the response. All other
              fields will be excluded.
            type: array
            collectionFormat: csv
            items:
              type: string
          - name: exclude_fields
            in: query
            required: false
            description: Specify fields to exclude from the response
            type: array
            collectionFormat: csv
            items:
              type: string
          - name: dereference_nested
            in: query
            required: false
            description: Commands and instances will be an object id
            type: boolean
            default: true
          - name: include_commands
            in: query
            required: false
            description: __DEPRECATED__ Include commands in the response.
              Use `exclude_fields=commands` instead.
            type: boolean
            default: true
        responses:
          200:
            description: All Systems
            schema:
              type: array
              items:
                $ref: '#/definitions/System'
          50x:
            $ref: '#/definitions/50xError'
        tags:
          - Systems
        """
        query_set = System.objects.order_by(
            self.request.headers.get("order_by", "name")
        )
        serialize_params = {"to_string": True, "many": True}

        include_fields = self.get_query_argument("include_fields", None)
        exclude_fields = self.get_query_argument("exclude_fields", None)
        dereference_nested = self.get_query_argument("dereference_nested", None)
        include_commands = self.get_query_argument("include_commands", None)

        if include_fields:
            include_fields = set(include_fields.split(",")) & self.REQUEST_FIELDS
            query_set = query_set.only(*include_fields)
            serialize_params["only"] = include_fields

        if exclude_fields:
            exclude_fields = set(exclude_fields.split(",")) & self.REQUEST_FIELDS
            query_set = query_set.exclude(*exclude_fields)
            serialize_params["exclude"] = exclude_fields

        if include_commands and include_commands.lower() == "false":
            query_set = query_set.exclude("commands")

            if "exclude" not in serialize_params:
                serialize_params["exclude"] = set()
            serialize_params["exclude"].add("commands")

        if dereference_nested and dereference_nested.lower() == "false":
            query_set = query_set.no_dereference()

        # TODO - Handle multiple query arguments with the same key
        # for example: (?name=foo&name=bar) ... what should that mean?
        filter_params = {}

        # Need to use self.request.query_arguments to get all the query args
        for key in self.request.query_arguments:
            if key in self.REQUEST_FIELDS:
                # Now use get_query_argument to get the decoded value
                filter_params[key] = self.get_query_argument(key)

        result_set = query_set.filter(**filter_params)

        self.set_header("Content-Type", "application/json; charset=UTF-8")
        self.write(self.parser.serialize_system(result_set, **serialize_params))

    @coroutine
    @authenticated(permissions=[Permissions.SYSTEM_CREATE])
    def post(self):
        """
        ---
        summary: Create a new System or update an existing System
        description: |
            If the System does not exist it will be created. If the System
            already exists it will be updated (assuming it passes validation).
        parameters:
          - name: system
            in: body
            description: The System definition to create / update
            schema:
              $ref: '#/definitions/System'
        responses:
          200:
            description: An existing System has been updated
            schema:
              $ref: '#/definitions/System'
          201:
            description: A new System has been created
            schema:
              $ref: '#/definitions/System'
          400:
            $ref: '#/definitions/400Error'
          50x:
            $ref: '#/definitions/50xError'
        tags:
          - Systems
        """
        self.request.event.name = Events.SYSTEM_CREATED.name

        system_model = self.parser.parse_system(
            self.request.decoded_body, from_string=True
        )

        with (yield self.system_lock.acquire()):
            # See if we already have a system with this name + version
            existing_system = System.find_unique(
                system_model.name, system_model.version
            )

            if not existing_system:
                self.logger.debug("Creating a new system: %s" % system_model.name)
                saved_system, status_code = self._create_new_system(system_model)
            else:
                self.logger.debug(
                    "System %s already exists. Updating it." % system_model.name
                )
                self.request.event.name = Events.SYSTEM_UPDATED.name
                saved_system, status_code = self._update_existing_system(
                    existing_system, system_model
                )

            saved_system.deep_save()

        self.request.event_extras = {"system": saved_system}

        self.set_status(status_code)
        self.write(
            self.parser.serialize_system(
                saved_system, to_string=False, include_commands=True
            )
        )

    @staticmethod
    def _create_new_system(system_model):
        new_system = system_model

        # Assign a default 'main' instance if there aren't any instances and there can
        # only be one
        if not new_system.instances or len(new_system.instances) == 0:
            if new_system.max_instances is None or new_system.max_instances == 1:
                new_system.instances = [Instance(name="default")]
                new_system.max_instances = 1
            else:
                raise ModelValidationError(
                    "Could not create system %s-%s: Systems with "
                    "max_instances > 1 must also define their instances"
                    % (system_model.name, system_model.version)
                )
        else:
            if not new_system.max_instances:
                new_system.max_instances = len(new_system.instances)

        return new_system, 201

    @staticmethod
    def _update_existing_system(existing_system, system_model):
        # Raise an exception if commands already exist for this system and they differ
        # from what's already in the database in a significant way
        if (
            existing_system.commands
            and "dev" not in existing_system.version
            and existing_system.has_different_commands(system_model.commands)
        ):
            raise ModelValidationError(
                "System %s-%s already exists with different commands"
                % (system_model.name, system_model.version)
            )
        else:
            existing_system.upsert_commands(system_model.commands)

        # Update instances
        if not system_model.instances or len(system_model.instances) == 0:
            system_model.instances = [Instance(name="default")]

        for attr in ["description", "icon_name", "display_name"]:
            value = getattr(system_model, attr)

            # If we set an attribute on the model as None, mongoengine marks the
            # attribute for deletion. We want to prevent this, so we set it to an emtpy
            # string.
            if value is None:
                setattr(existing_system, attr, "")
            else:
                setattr(existing_system, attr, value)

        # Update metadata
        new_metadata = system_model.metadata or {}
        existing_system.metadata.update(new_metadata)

        old_instances = [
            inst
            for inst in existing_system.instances
            if system_model.has_instance(inst.name)
        ]
        new_instances = [
            inst
            for inst in system_model.instances
            if not existing_system.has_instance(inst.name)
        ]
        existing_system.instances = old_instances + new_instances

        return existing_system, 200



import logging

from tornado.gen import coroutine

from bg_utils.mongo.parser import MongoParser
from brew_view import thrift_context
from brew_view.authorization import check_permission, Permissions
from brew_view.base_handler import BaseHandler
from brewtils.errors import ModelValidationError


class AdminAPI(BaseHandler):

    parser = MongoParser()
    logger = logging.getLogger(__name__)

    @coroutine
    def patch(self):
        """
        ---
        summary: Initiate a rescan of the plugin directory
        description: |
          The body of the request needs to contain a set of instructions
          detailing the operations to perform.

          Currently the only operation supported is `rescan`:
          ```JSON
          {
            "operations": [
              { "operation": "rescan" }
            ]
          }
          ```
          * Will remove from the registry and database any currently stopped
            plugins who's directory has been removed.
          * Will add and start any new plugin directories.
        parameters:
          - name: patch
            in: body
            required: true
            description: Instructions for operations
            schema:
              $ref: '#/definitions/Patch'
        responses:
          204:
            description: Rescan successfully initiated
          50x:
            $ref: '#/definitions/50xError'
        tags:
          - Admin
        """
        operations = self.parser.parse_patch(
            self.request.decoded_body, many=True, from_string=True
        )

        for op in operations:
            if op.operation == "rescan":
                check_permission(self.current_user, [Permissions.SYSTEM_CREATE])
                with thrift_context() as client:
                    yield client.rescanSystemDirectory()
            else:
                error_msg = "Unsupported operation '%s'" % op.operation
                self.logger.warning(error_msg)
                raise ModelValidationError(error_msg)

        self.set_status(204)


class OldAdminAPI(BaseHandler):
    @coroutine
    def post(self):
        """
        ---
        summary: Initiate a rescan of the plugin directory
        deprecated: true
        description: |
          This endpoint is DEPRECATED - Use PATCH /api/v1/admin instead.

          Will initiate a rescan of the plugins directory.
          * Will remove from the registry and database any currently stopped
            plugins who's directory has been removed.
          * Will add and start any new plugin directories.
        responses:
          204:
            description: Rescan successfully initiated
          50x:
            $ref: '#/definitions/50xError'
        tags:
          - Deprecated
        """
        with thrift_context() as client:
            yield client.rescanSystemDirectory()

        self.set_status(204)



import json
import logging

from brew_view.authorization import authenticated, Permissions
from brew_view.base_handler import BaseHandler


class PermissionsAPI(BaseHandler):

    logger = logging.getLogger(__name__)

    @authenticated(permissions=[Permissions.USER_READ])
    def get(self):
        """
        ---
        summary: Retrieve all Permissions
        responses:
          200:
            description: All Permissions
            schema:
              type: array
              items:
                type: string
          50x:
            $ref: '#/definitions/50xError'
        tags:
          - Permissions
        """
        self.set_header("Content-Type", "application/json; charset=UTF-8")
        self.write(json.dumps(sorted(Permissions.values)))



import json
import logging

from mongoengine.errors import DoesNotExist, ValidationError
from passlib.apps import custom_app_context

import brew_view
from bg_utils.mongo.models import Principal, Role
from bg_utils.mongo.parser import MongoParser
from brew_view.authorization import (
    authenticated,
    check_permission,
    Permissions,
    coalesce_permissions,
)
from brew_view.base_handler import BaseHandler
from brewtils.errors import ModelValidationError


class UserAPI(BaseHandler):

    logger = logging.getLogger(__name__)

    def get(self, user_identifier):
        """
        ---
        summary: Retrieve a specific User
        parameters:
          - name: user_identifier
            in: path
            required: true
            description: The ID or name of the User
            type: string
        responses:
          200:
            description: User with the given ID
            schema:
              $ref: '#/definitions/User'
          404:
            $ref: '#/definitions/404Error'
          50x:
            $ref: '#/definitions/50xError'
        tags:
          - Users
        """
        if user_identifier == "anonymous":
            principal = brew_view.anonymous_principal
        else:
            # Need fine-grained access control here
            if user_identifier not in [
                self.current_user.id,
                self.current_user.username,
            ]:
                check_permission(self.current_user, [Permissions.USER_READ])

            try:
                principal = Principal.objects.get(id=str(user_identifier))
            except (DoesNotExist, ValidationError):
                principal = Principal.objects.get(username=str(user_identifier))

        principal.permissions = coalesce_permissions(principal.roles)[1]

        self.write(MongoParser.serialize_principal(principal, to_string=False))

    @authenticated(permissions=[Permissions.USER_DELETE])
    def delete(self, user_id):
        """
        ---
        summary: Delete a specific User
        parameters:
          - name: user_id
            in: path
            required: true
            description: The ID of the User
            type: string
        responses:
          204:
            description: User has been successfully deleted
          404:
            $ref: '#/definitions/404Error'
          50x:
            $ref: '#/definitions/50xError'
        tags:
          - Users
        """
        principal = Principal.objects.get(id=str(user_id))
        principal.delete()

        self.set_status(204)

    def patch(self, user_id):
        """
        ---
        summary: Partially update a User
        description: |
          The body of the request needs to contain a set of instructions
          detailing the updates to apply:
          ```JSON
          {
            "operations": [
              { "operation": "add", "path": "/roles", "value": "admin" }
            ]
          }
          ```
        parameters:
          - name: user_id
            in: path
            required: true
            description: The ID of the User
            type: string
          - name: patch
            in: body
            required: true
            description: Instructions for how to update the User
            schema:
              $ref: '#/definitions/Patch'
        responses:
          200:
            description: User with the given ID
            schema:
              $ref: '#/definitions/User'
          400:
            $ref: '#/definitions/400Error'
          404:
            $ref: '#/definitions/404Error'
          50x:
            $ref: '#/definitions/50xError'
        tags:
          - Users
        """
        principal = Principal.objects.get(id=str(user_id))
        operations = MongoParser.parse_patch(
            self.request.decoded_body, many=True, from_string=True
        )

        for op in operations:
            if op.path == "/roles":
                check_permission(self.current_user, [Permissions.USER_UPDATE])

                try:
                    if op.operation == "add":
                        principal.roles.append(Role.objects.get(name=op.value))
                    elif op.operation == "remove":
                        principal.roles.remove(Role.objects.get(name=op.value))
                    elif op.operation == "set":
                        principal.roles = [
                            Role.objects.get(name=name) for name in op.value
                        ]
                    else:
                        raise ModelValidationError(
                            "Unsupported operation '%s'" % op.operation
                        )
                except DoesNotExist:
                    raise ModelValidationError("Role '%s' does not exist" % op.value)

            elif op.path == "/preferences/theme":
                if user_id != self.current_user.id:
                    check_permission(self.current_user, [Permissions.USER_UPDATE])

                if op.operation == "set":
                    principal.preferences["theme"] = op.value
                else:
                    raise ModelValidationError(
                        "Unsupported operation '%s'" % op.operation
                    )

            else:
                check_permission(self.current_user, [Permissions.USER_UPDATE])
                raise ModelValidationError("Unsupported path '%s'" % op.path)

        principal.save()

        self.write(MongoParser.serialize_principal(principal, to_string=False))


class UsersAPI(BaseHandler):
    @authenticated(permissions=[Permissions.USER_READ])
    def get(self):
        """
        ---
        summary: Retrieve all Users
        responses:
          200:
            description: All Users
            schema:
              type: array
              items:
                $ref: '#/definitions/User'
          50x:
            $ref: '#/definitions/50xError'
        tags:
          - Users
        """
        principals = Principal.objects.all().select_related(max_depth=1)

        for principal in principals:
            principal.permissions = coalesce_permissions(principal.roles)[1]

        self.set_header("Content-Type", "application/json; charset=UTF-8")
        self.write(
            MongoParser.serialize_principal(principals, to_string=True, many=True)
        )

    @authenticated(permissions=[Permissions.USER_CREATE])
    def post(self):
        """
        ---
        summary: Create a new User
        parameters:
          - name: user
            in: body
            description: The user
            schema:
              type: object
              properties:
                username:
                  type: string
                  description: the name
                password:
                  type: string
                  description: the password
              required:
                - username
                - password
        consumes:
          - application/json
        responses:
          201:
            description: A new User has been created
            schema:
              $ref: '#/definitions/User'
          400:
            $ref: '#/definitions/400Error'
          50x:
            $ref: '#/definitions/50xError'
        tags:
          - Users
        """
        parsed = json.loads(self.request.decoded_body)

        user = Principal(
            username=parsed["username"],
            hash=custom_app_context.hash(parsed["password"]),
        )
        user.save()

        self.set_status(204)



#!/usr/bin/env python

import signal
import sys
from argparse import ArgumentParser

from yapconf import YapconfSpec

import bg_utils
import brew_view
from brew_view.specification import SPECIFICATION, get_default_logging_config


def signal_handler(signal_number, stack_frame):
    brew_view.logger.info("Received a shutdown request.")
    brew_view.io_loop.add_callback_from_signal(brew_view.shutdown)


def generate_logging_config():
    spec = YapconfSpec(SPECIFICATION, env_prefix="BG_")
    bg_utils.generate_logging_config_file(
        spec, get_default_logging_config, sys.argv[1:]
    )


def generate_config():
    spec = YapconfSpec(SPECIFICATION, env_prefix="BG_")
    bg_utils.generate_config_file(spec, sys.argv[1:])


def migrate_config():
    spec = YapconfSpec(SPECIFICATION, env_prefix="BG_")
    bg_utils.update_config_file(spec, sys.argv[1:])


def main():
    signal.signal(signal.SIGINT, signal_handler)
    signal.signal(signal.SIGTERM, signal_handler)

    spec = YapconfSpec(SPECIFICATION, env_prefix="BG_")
    parser = ArgumentParser()
    spec.add_arguments(parser)
    args = parser.parse_args(sys.argv[1:])

    # Logging isn't set up until after this...
    brew_view.setup(spec, vars(args))

    # Schedule things to happen after the ioloop comes up
    brew_view.io_loop.add_callback(brew_view.startup)

    brew_view.logger.info("Starting IO loop")
    brew_view.io_loop.start()

    brew_view.logger.info("Application is shut down. Goodbye!")


if __name__ == "__main__":
    main()



import logging

from tornado.web import HTTPError
from tornado.websocket import WebSocketHandler

import brew_view
from bg_utils.mongo.parser import MongoParser
from brew_view.authorization import (
    authenticated,
    check_permission,
    query_token_auth,
    AuthMixin,
    Permissions,
)
from brew_view.base_handler import BaseHandler
from brewtils.errors import RequestForbidden
from brewtils.schema_parser import SchemaParser


class EventPublisherAPI(BaseHandler):

    logger = logging.getLogger(__name__)
    parser = SchemaParser()

    @authenticated(permissions=[Permissions.EVENT_CREATE])
    def post(self):
        """
        ---
        summary: Publish a new event
        parameters:
          - name: event
            in: body
            description: The the Event object
            schema:
              $ref: '#/definitions/Event'
          - name: publisher
            in: query
            required: false
            description: Specific publisher to use
            type: array
            collectionFormat: multi
            items:
              properties:
                data:
                  type: string
        responses:
          204:
            description: An Event has been published
          400:
            $ref: '#/definitions/400Error'
          50x:
            $ref: '#/definitions/50xError'
        tags:
          - Beta
        """
        event = self.parser.parse_event(self.request.decoded_body, from_string=True)
        publishers = self.get_query_arguments("publisher")

        if not publishers:
            brew_view.event_publishers.publish_event(event)
        else:
            for publisher in publishers:
                brew_view.event_publishers[publisher].publish_event(event)

        self.set_status(204)


class EventSocket(AuthMixin, WebSocketHandler):

    logger = logging.getLogger(__name__)
    parser = MongoParser()

    closing = False
    listeners = set()

    def __init__(self, *args, **kwargs):
        super(EventSocket, self).__init__(*args, **kwargs)

        self.auth_providers.append(query_token_auth)

    def check_origin(self, origin):
        return True

    def open(self):
        if EventSocket.closing:
            self.close(reason="Shutting down")
            return

        # We can't go though the 'normal' BaseHandler exception translation
        try:
            check_permission(self.current_user, [Permissions.EVENT_READ])
        except (HTTPError, RequestForbidden) as ex:
            self.close(reason=str(ex))
            return

        EventSocket.listeners.add(self)

    def on_close(self):
        EventSocket.listeners.discard(self)

    def on_message(self, message):
        pass

    @classmethod
    def publish(cls, message):
        # Don't bother if nobody is listening
        if not len(cls.listeners):
            return

        for listener in cls.listeners:
            listener.write_message(message)

    @classmethod
    def shutdown(cls):
        cls.logger.debug("Closing websocket connections")
        EventSocket.closing = True

        for listener in cls.listeners:
            listener.close(reason="Shutting down")



# -*- coding: utf-8 -*-
"""Module for the job runner."""
import brew_view


def run_job(job_id, request_template):
    """Spawned by the scheduler, this will kick off a new request.

    This method is meant to be run in a separate process.

    Args:
        job_id: The Beer-Garden job ID that triggered this event.
        request_template: Request template specified by the job.
    """
    request_template.metadata["_bg_job_id"] = job_id
    brew_view.easy_client.create_request(request_template)



import logging

from tornado.gen import coroutine

from brew_view import thrift_context
from brew_view.authorization import authenticated, Permissions
from brew_view.base_handler import BaseHandler
from brewtils.models import Events


class QueueAPI(BaseHandler):

    logger = logging.getLogger(__name__)

    @coroutine
    @authenticated(permissions=[Permissions.QUEUE_DELETE])
    def delete(self, queue_name):
        """
        ---
        summary: Clear a queue by canceling all requests
        parameters:
          - name: queue_name
            in: path
            required: true
            description: The name of the queue to clear
            type: string
        responses:
          204:
            description: Queue successfully cleared
          404:
            $ref: '#/definitions/404Error'
          50x:
            $ref: '#/definitions/50xError'
        tags:
          - Queues
        """
        self.request.event.name = Events.QUEUE_CLEARED.name
        self.request.event.payload = {"queue_name": queue_name}

        with thrift_context() as client:
            yield client.clearQueue(queue_name)

        self.set_status(204)


class OldQueueAPI(BaseHandler):

    logger = logging.getLogger(__name__)

    def delete(self, queue_name):
        """
        ---
        summary: Clear a queue by canceling all requests
        deprecated: true
        description: This endpoint is DEPRECATED - Use /api/v1/queues/{queue_name}
            instead.
        parameters:
          - name: queue_name
            in: path
            required: true
            description: The name of the queue to clear
            type: string
        responses:
          204:
            description: Queue successfully cleared
          404:
            $ref: '#/definitions/404Error'
          50x:
            $ref: '#/definitions/50xError'
        tags:
          - Deprecated
        """
        self.redirect("/api/v1/queues/" + queue_name, permanent=True)



import datetime
import re
import socket

from jwt.exceptions import ExpiredSignatureError, InvalidSignatureError
from mongoengine.errors import (
    DoesNotExist,
    NotUniqueError,
    ValidationError as MongoValidationError,
)
from thriftpy.thrift import TException
from tornado.web import HTTPError, RequestHandler

import bg_utils
import brew_view
from brew_view.authorization import AuthMixin
from brew_view.metrics import http_api_latency_total, request_latency
from brewtils.errors import (
    ModelError,
    ModelValidationError,
    RequestForbidden,
    RequestPublishException,
    WaitExceededError,
    AuthorizationRequired,
)
from brewtils.models import Event


class BaseHandler(AuthMixin, RequestHandler):
    """Base handler from which all handlers inherit"""

    MONGO_ID_PATTERN = r".*/([0-9a-f]{24}).*"

    charset_re = re.compile(r"charset=(.*)$")

    error_map = {
        MongoValidationError: {"status_code": 400},
        ModelError: {"status_code": 400},
        bg_utils.bg_thrift.InvalidSystem: {"status_code": 400},
        ExpiredSignatureError: {"status_code": 401},
        AuthorizationRequired: {"status_code": 401},
        RequestForbidden: {"status_code": 403},
        InvalidSignatureError: {"status_code": 403},
        DoesNotExist: {"status_code": 404, "message": "Resource does not exist"},
        WaitExceededError: {"status_code": 408, "message": "Max wait time exceeded"},
        NotUniqueError: {"status_code": 409, "message": "Resource already exists"},
        RequestPublishException: {"status_code": 502},
        bg_utils.bg_thrift.BaseException: {
            "status_code": 502,
            "message": "An error occurred " "on the backend",
        },
        TException: {"status_code": 503, "message": "Could not connect to Bartender"},
        socket.timeout: {"status_code": 504, "message": "Backend request timed out"},
    }

    def set_default_headers(self):
        """Headers set here will be applied to all responses"""
        self.set_header("BG-Version", brew_view.__version__)

        if brew_view.config.cors_enabled:
            self.set_header("Access-Control-Allow-Origin", "*")
            self.set_header("Access-Control-Allow-Headers", "Content-Type")
            self.set_header(
                "Access-Control-Allow-Methods", "GET, POST, PATCH, DELETE, OPTIONS"
            )

    @property
    def prometheus_endpoint(self):
        """Removes Mongo ID from endpoint."""
        to_return = self.request.path.rstrip("/")
        for mongo_id in re.findall(self.MONGO_ID_PATTERN, self.request.path):
            to_return = to_return.replace(mongo_id, "<ID>")
        return to_return

    def prepare(self):
        """Called before each verb handler"""

        # Used for calculating request handling duration
        self.request.created_time = datetime.datetime.utcnow()

        # This is used for sending event notifications
        self.request.event = Event()
        self.request.event_extras = {}

        content_type = self.request.headers.get("content-type", "")
        if self.request.method.upper() in ["POST", "PATCH"] and content_type:
            content_type = content_type.split(";")

            self.request.mime_type = content_type[0]
            if self.request.mime_type not in [
                "application/json",
                "application/x-www-form-urlencoded",
            ]:
                raise ModelValidationError("Unsupported or missing content-type header")

            # Attempt to parse out the charset and decode the body, default to utf-8
            charset = "utf-8"
            if len(content_type) > 1:
                search_result = self.charset_re.search(content_type[1])
                if search_result:
                    charset = search_result.group(1)
            self.request.charset = charset
            self.request.decoded_body = self.request.body.decode(charset)

    def on_finish(self):
        """Called after a handler completes processing"""
        # This is gross, but in some cases we have to do these in the handler
        if getattr(self.request, "publish_metrics", True):
            http_api_latency_total.labels(
                method=self.request.method.upper(),
                route=self.prometheus_endpoint,
                status=self.get_status(),
            ).observe(request_latency(self.request.created_time))

        if self.request.event.name and getattr(self.request, "publish_event", True):
            brew_view.event_publishers.publish_event(
                self.request.event, **self.request.event_extras
            )

    def options(self, *args, **kwargs):

        if brew_view.config.cors_enabled:
            self.set_status(204)
        else:
            raise HTTPError(403, reason="CORS is disabled")

    def write_error(self, status_code, **kwargs):
        """Transform an exception into a response.

        This protects controllers from having to write a lot of the same code over and
        over and over. Controllers can, of course, overwrite error handlers and return
        their own responses if necessary, but generally, this is where error handling
        should occur.

        When an exception is handled this function makes two passes through error_map.
        The first pass is to see if the exception type can be matched exactly. If there
        is no exact type match the second pass will attempt to match using isinstance.
        If a message is provided in the error_map it takes precedence over the
        exception message.

        ***NOTE*** Nontrivial inheritance trees will almost definitely break. This is a
        BEST EFFORT using a simple isinstance check on an unordered data structure. So
        if an exception class has both a parent and a grandparent in the error_map
        there is no guarantee about which message / status code will be chosen. The
        same applies to exceptions that use multiple inheritance.

        ***LOGGING***
        An exception raised in a controller method will generate logging to the
        tornado.application logger that includes a stacktrace. That logging occurs
        before this method is invoked. The result of this method will generate logging
        to the tornado.access logger as usual. So there is no need to do additional
        logging here as the 'real' exception will already have been logged.

        :param status_code: a status_code that will be used if no match is found in the
        error map
        :return: None
        """
        code = 0
        message = ""

        if "exc_info" in kwargs:
            typ3 = kwargs["exc_info"][0]
            e = kwargs["exc_info"][1]

            error_dict = None
            if typ3 in self.error_map.keys():
                error_dict = self.error_map[typ3]
            else:
                for error_type in self.error_map.keys():
                    if isinstance(e, error_type):
                        error_dict = self.error_map[error_type]
                        break

            if error_dict:
                code = error_dict.get("status_code", 500)
                message = error_dict.get("message", str(e))

            elif brew_view.config.debug_mode:
                message = str(e)

        code = code or status_code or 500
        message = message or (
            "Encountered unknown exception. Please check "
            "with your System Administrator."
        )

        self.request.event.error = True
        self.request.event.payload = {"message": message}

        self.set_header("Content-Type", "application/json; charset=UTF-8")
        self.set_status(code)
        self.finish({"message": message})



import contextlib
import logging
import os
from concurrent.futures import ThreadPoolExecutor

import ssl
from apispec import APISpec
from apscheduler.executors.tornado import TornadoExecutor
from apscheduler.schedulers.tornado import TornadoScheduler
from functools import partial
from prometheus_client.exposition import start_http_server
from pytz import utc
from thriftpy.rpc import client_context
from tornado.concurrent import Future
from tornado.gen import coroutine, sleep
from tornado.httpserver import HTTPServer
from tornado.ioloop import IOLoop
from tornado.web import Application, StaticFileHandler, RedirectHandler
from urllib3.util.url import Url

import bg_utils
import brew_view._version
from bg_utils.event_publisher import EventPublishers
from bg_utils.mongo import setup_database
from bg_utils.pika import TransientPikaClient
from bg_utils.plugin_logging_loader import PluginLoggingLoader
from brew_view.authorization import anonymous_principal as load_anonymous
from brew_view.metrics import initialize_counts
from brew_view.publishers import (
    MongoPublisher,
    RequestPublisher,
    TornadoPikaPublisher,
    WebsocketPublisher,
)
from brew_view.scheduler.jobstore import BGJobStore
from brew_view.specification import get_default_logging_config
from brewtils.models import Event, Events
from brewtils.rest import normalize_url_prefix
from brewtils.rest.easy_client import EasyClient
from brewtils.schemas import (
    ParameterSchema,
    CommandSchema,
    InstanceSchema,
    SystemSchema,
    RequestSchema,
    PatchSchema,
    LoggingConfigSchema,
    EventSchema,
    QueueSchema,
    PrincipalSchema,
    RoleSchema,
    RefreshTokenSchema,
    JobSchema,
    DateTriggerSchema,
    IntervalTriggerSchema,
    CronTriggerSchema,
)

__version__ = brew_view._version.__version__

config = None
io_loop = None
server = None
tornado_app = None
public_url = None
logger = None
thrift_context = None
event_publishers = None
api_spec = None
plugin_logging_config = None
app_logging_config = None
notification_meta = None
request_scheduler = None
request_map = {}
anonymous_principal = None
easy_client = None
client_ssl = None


def setup(spec, cli_args):
    global config, logger, app_logging_config

    config = bg_utils.load_application_config(spec, cli_args)

    app_logging_config = bg_utils.setup_application_logging(
        config, get_default_logging_config(config.log.level, config.log.file)
    )
    logger = logging.getLogger(__name__)
    logger.debug("Logging configured. First post!")

    load_plugin_logging_config(config)
    _setup_application()


@coroutine
def startup():
    """Do startup things.

    This is the first thing called from within the ioloop context.
    """
    global event_publishers, anonymous_principal

    # Ensure we have a mongo connection
    logger.info("Checking for Mongo connection")
    yield _progressive_backoff(
        partial(setup_database, config), "Unable to connect to mongo, is it started?"
    )

    # Need to wait until after mongo connection established to load
    anonymous_principal = load_anonymous()

    logger.info("Starting event publishers")
    event_publishers = _setup_event_publishers(client_ssl)

    logger.info("Initializing metrics")
    initialize_counts()

    logger.info(
        "Starting metrics server on %s:%d" % (config.web.host, config.metrics.port)
    )
    start_http_server(config.metrics.port)

    logger.info("Starting HTTP server on %s:%d" % (config.web.host, config.web.port))
    server.listen(config.web.port, config.web.host)

    logger.info("Starting scheduler")
    request_scheduler.start()

    logger.debug("Publishing application startup event")
    event_publishers.publish_event(Event(name=Events.BREWVIEW_STARTED.name))

    brew_view.logger.info("Application is started. Hello!")


@coroutine
def shutdown():
    """Do shutdown things

    This still operates within the ioloop, so stopping it should be the last
    thing done.

    Because things in startup aren't guaranteed to have been run we need to be
    careful about checking to make sure things actually need to be shut down.

    This execution is normally scheduled by the signal handler.
    """
    if request_scheduler.running:
        logger.info("Stopping scheduler")
        request_scheduler.shutdown(wait=False)

    logger.info("Stopping HTTP server")
    server.stop()

    if event_publishers:
        logger.debug("Publishing application shutdown event")
        event_publishers.publish_event(Event(name=Events.BREWVIEW_STOPPED.name))

        logger.info("Shutting down event publishers")
        yield list(filter(lambda x: isinstance(x, Future), event_publishers.shutdown()))

    logger.info("Stopping IO loop")
    io_loop.add_callback(io_loop.stop)


@coroutine
def _progressive_backoff(func, failure_message):
    wait_time = 0.1
    while not func():
        logger.warning(failure_message)
        logger.warning("Waiting %.1f seconds before next attempt", wait_time)

        yield sleep(wait_time)
        wait_time = min(wait_time * 2, 30)


def load_plugin_logging_config(input_config):
    global plugin_logging_config

    plugin_logging_config = PluginLoggingLoader().load(
        filename=input_config.plugin_logging.config_file,
        level=input_config.plugin_logging.level,
        default_config=app_logging_config,
    )


def _setup_application():
    """Setup things that can be taken care of before io loop is started"""
    global io_loop, tornado_app, public_url, thrift_context, easy_client
    global server, client_ssl, request_scheduler

    # Tweak some config options
    config.web.url_prefix = normalize_url_prefix(config.web.url_prefix)
    if not config.auth.token.secret:
        config.auth.token.secret = os.urandom(20)
        if config.auth.enabled:
            logger.warning(
                "Brew-view was started with authentication enabled and no "
                "Secret. Generated tokens will not be valid across Brew-view "
                "restarts. To prevent this set the auth.token.secret config."
            )

    public_url = Url(
        scheme="https" if config.web.ssl.enabled else "http",
        host=config.web.public_fqdn,
        port=config.web.port,
        path=config.web.url_prefix,
    ).url

    # This is not super clean as we're pulling the config from different
    # 'sections,' but the scheduler is the only thing that uses this
    easy_client = EasyClient(
        host=config.web.public_fqdn,
        port=config.web.port,
        url_prefix=config.web.url_prefix,
        ssl_enabled=config.web.ssl.enabled,
        ca_cert=config.web.ssl.ca_cert,
        username=config.scheduler.auth.username,
        password=config.scheduler.auth.password,
    )

    thrift_context = _setup_thrift_context()
    tornado_app = _setup_tornado_app()
    server_ssl, client_ssl = _setup_ssl_context()
    request_scheduler = _setup_scheduler()

    server = HTTPServer(tornado_app, ssl_options=server_ssl)
    io_loop = IOLoop.current()


def _setup_scheduler():
    jobstores = {"beer_garden": BGJobStore()}
    # TODO: Look at creating a custom executor using process pools
    executors = {"default": TornadoExecutor(config.scheduler.max_workers)}
    job_defaults = config.scheduler.job_defaults.to_dict()

    return TornadoScheduler(
        jobstores=jobstores,
        executors=executors,
        job_defaults=job_defaults,
        timezone=utc,
    )


def _setup_tornado_app():

    # Import these here so we don't have a problem importing thrift_context
    from brew_view.controllers import (
        AdminAPI,
        CommandAPI,
        CommandListAPI,
        ConfigHandler,
        InstanceAPI,
        QueueAPI,
        QueueListAPI,
        RequestAPI,
        RequestListAPI,
        SystemAPI,
        SystemListAPI,
        VersionHandler,
        SpecHandler,
        SwaggerConfigHandler,
        OldAdminAPI,
        OldQueueAPI,
        OldQueueListAPI,
        LoggingConfigAPI,
        EventPublisherAPI,
        EventSocket,
        TokenAPI,
        UserAPI,
        UsersAPI,
        RoleAPI,
        RolesAPI,
        TokenListAPI,
        JobAPI,
        JobListAPI,
        PermissionsAPI,
    )

    prefix = config.web.url_prefix
    static_base = os.path.join(os.path.dirname(__file__), "static", "dist")

    # These get documented in our OpenAPI (fka Swagger) documentation
    published_url_specs = [
        (r"{0}api/v1/commands/?".format(prefix), CommandListAPI),
        (r"{0}api/v1/requests/?".format(prefix), RequestListAPI),
        (r"{0}api/v1/systems/?".format(prefix), SystemListAPI),
        (r"{0}api/v1/queues/?".format(prefix), QueueListAPI),
        (r"{0}api/v1/users/?".format(prefix), UsersAPI),
        (r"{0}api/v1/roles/?".format(prefix), RolesAPI),
        (r"{0}api/v1/permissions/?".format(prefix), PermissionsAPI),
        (r"{0}api/v1/tokens/?".format(prefix), TokenListAPI),
        (r"{0}api/v1/admin/?".format(prefix), AdminAPI),
        (r"{0}api/v1/jobs/?".format(prefix), JobListAPI),
        (r"{0}api/v1/commands/(\w+)/?".format(prefix), CommandAPI),
        (r"{0}api/v1/instances/(\w+)/?".format(prefix), InstanceAPI),
        (r"{0}api/v1/requests/(\w+)/?".format(prefix), RequestAPI),
        (r"{0}api/v1/systems/(\w+)/?".format(prefix), SystemAPI),
        (r"{0}api/v1/queues/([\w\.-]+)/?".format(prefix), QueueAPI),
        (r"{0}api/v1/users/(\w+)/?".format(prefix), UserAPI),
        (r"{0}api/v1/roles/(\w+)/?".format(prefix), RoleAPI),
        (r"{0}api/v1/tokens/(\w+)/?".format(prefix), TokenAPI),
        (r"{0}api/v1/jobs/(\w+)/?".format(prefix), JobAPI),
        (r"{0}api/v1/config/logging/?".format(prefix), LoggingConfigAPI),
        # Beta
        (r"{0}api/vbeta/events/?".format(prefix), EventPublisherAPI),
        # Deprecated
        (r"{0}api/v1/admin/system/?".format(prefix), OldAdminAPI),
        (r"{0}api/v1/admin/queues/?".format(prefix), OldQueueListAPI),
        (r"{0}api/v1/admin/queues/([\w\.-]+)/?".format(prefix), OldQueueAPI),
    ]

    # And these do not
    unpublished_url_specs = [
        # These are a little special - unpublished but still versioned
        # The swagger spec
        (r"{0}api/v1/spec/?".format(prefix), SpecHandler),
        # Events websocket
        (r"{0}api/v1/socket/events/?".format(prefix), EventSocket),
        # Version / configs
        (r"{0}version/?".format(prefix), VersionHandler),
        (r"{0}config/?".format(prefix), ConfigHandler),
        (r"{0}config/swagger/?".format(prefix), SwaggerConfigHandler),
        # Not sure if these are really necessary
        (r"{0}".format(prefix[:-1]), RedirectHandler, {"url": prefix}),
        (
            r"{0}swagger/(.*)".format(prefix),
            StaticFileHandler,
            {"path": os.path.join(static_base, "swagger")},
        ),
        # Static content
        (
            r"{0}(.*)".format(prefix),
            StaticFileHandler,
            {"path": static_base, "default_filename": "index.html"},
        ),
    ]
    _load_swagger(published_url_specs, title=config.application.name)

    return Application(
        published_url_specs + unpublished_url_specs, debug=config.debug_mode
    )


def _setup_ssl_context():

    if config.web.ssl.enabled:
        server_ssl = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
        server_ssl.load_cert_chain(
            certfile=config.web.ssl.public_key, keyfile=config.web.ssl.private_key
        )
        server_ssl.verify_mode = getattr(
            ssl, "CERT_" + config.web.ssl.client_cert_verify.upper()
        )

        client_ssl = ssl.create_default_context(ssl.Purpose.SERVER_AUTH)
        client_ssl.load_cert_chain(
            certfile=config.web.ssl.public_key, keyfile=config.web.ssl.private_key
        )

        if config.web.ssl.ca_cert or config.web.ssl.ca_path:
            server_ssl.load_verify_locations(
                cafile=config.web.ssl.ca_cert, capath=config.web.ssl.ca_path
            )
            client_ssl.load_verify_locations(
                cafile=config.web.ssl.ca_cert, capath=config.web.ssl.ca_path
            )
    else:
        server_ssl = None
        client_ssl = None

    return server_ssl, client_ssl


def _setup_thrift_context():
    class BgClient(object):
        """Helper class that wraps a thriftpy TClient"""

        executor = ThreadPoolExecutor(max_workers=10)

        def __init__(self, t_client):
            self.t_client = t_client

        def __getattr__(self, thrift_method):
            def submit(*args, **kwargs):
                return self.executor.submit(
                    self.t_client.__getattr__(thrift_method), *args, **kwargs
                )

            return submit

    @contextlib.contextmanager
    def bg_thrift_context(sync=False, **kwargs):
        with client_context(
            bg_utils.bg_thrift.BartenderBackend,
            host=config.backend.host,
            port=config.backend.port,
            socket_timeout=config.backend.socket_timeout,
            **kwargs
        ) as client:
            yield client if sync else BgClient(client)

    return bg_thrift_context


def _setup_event_publishers(ssl_context):
    from brew_view.controllers.event_api import EventSocket

    # Create the collection of event publishers and add concrete publishers
    pubs = EventPublishers(
        {
            "request": RequestPublisher(ssl_context=ssl_context),
            "websocket": WebsocketPublisher(EventSocket),
        }
    )

    if config.event.mongo.enable:
        try:
            pubs["mongo"] = MongoPublisher()
        except Exception as ex:
            logger.warning("Error starting Mongo event publisher: %s", ex)

    if config.event.amq.enable:
        try:
            pika_params = {
                "host": config.amq.host,
                "port": config.amq.connections.message.port,
                "ssl": config.amq.connections.message.ssl,
                "user": config.amq.connections.admin.user,
                "password": config.amq.connections.admin.password,
                "exchange": config.event.amq.exchange,
                "virtual_host": config.event.amq.virtual_host,
                "connection_attempts": config.amq.connection_attempts,
            }

            # Make sure the exchange exists
            TransientPikaClient(**pika_params).declare_exchange()

            pubs["pika"] = TornadoPikaPublisher(
                shutdown_timeout=config.shutdown_timeout, **pika_params
            )
        except Exception as ex:
            logger.exception("Error starting RabbitMQ event publisher: %s", ex)

    # Metadata functions - additional metadata to be included with each event
    pubs.metadata_funcs["public_url"] = lambda: public_url

    return pubs


def _load_swagger(url_specs, title=None):

    global api_spec
    api_spec = APISpec(
        title=title,
        version="1.0",
        plugins=("apispec.ext.marshmallow", "apispec.ext.tornado"),
    )

    # Schemas from Marshmallow
    api_spec.definition("Parameter", schema=ParameterSchema)
    api_spec.definition("Command", schema=CommandSchema)
    api_spec.definition("Instance", schema=InstanceSchema)
    api_spec.definition("Request", schema=RequestSchema)
    api_spec.definition("System", schema=SystemSchema)
    api_spec.definition("LoggingConfig", schema=LoggingConfigSchema)
    api_spec.definition("Event", schema=EventSchema)
    api_spec.definition("User", schema=PrincipalSchema)
    api_spec.definition("Role", schema=RoleSchema)
    api_spec.definition("Queue", schema=QueueSchema)
    api_spec.definition("RefreshToken", schema=RefreshTokenSchema)
    api_spec.definition("_patch", schema=PatchSchema)
    api_spec.definition(
        "Patch",
        properties={
            "operations": {"type": "array", "items": {"$ref": "#/definitions/_patch"}}
        },
    )
    api_spec.definition("DateTrigger", schema=DateTriggerSchema)
    api_spec.definition("CronTrigger", schema=CronTriggerSchema)
    api_spec.definition("IntervalTrigger", schema=IntervalTriggerSchema)
    api_spec.definition("Job", schema=JobSchema)
    trigger_properties = {
        "allOf": [
            {"$ref": "#/definitions/CronTrigger"},
            {"$ref": "#/definitions/DateTrigger"},
            {"$ref": "#/definitions/IntervalTrigger"},
        ]
    }
    api_spec._definitions["Job"]["properties"]["trigger"] = trigger_properties

    error = {"message": {"type": "string"}}
    api_spec.definition(
        "400Error", properties=error, description="Parameter validation error"
    )
    api_spec.definition(
        "404Error", properties=error, description="Resource does not exist"
    )
    api_spec.definition(
        "409Error", properties=error, description="Resource already exists"
    )
    api_spec.definition("50xError", properties=error, description="Server exception")

    # Finally, add documentation for all our published paths
    for url_spec in url_specs:
        api_spec.add_path(urlspec=url_spec)



import logging
from datetime import timedelta
from functools import wraps

from pika import BasicProperties
from pika import TornadoConnection
from six import string_types
from tornado.concurrent import return_future
from tornado.gen import coroutine
from tornado.httpclient import AsyncHTTPClient
from tornado.ioloop import IOLoop
from tornado.queues import Queue

import brew_view
from bg_utils.event_publisher import EventPublisher
from bg_utils.mongo.parser import MongoParser
from bg_utils.pika import get_routing_key
from brewtils.models import Events
from brewtils.queues import PikaClient
from brewtils.rest.client import RestClient


class BeergardenPublisher(EventPublisher):
    def __init__(self):
        self._client = RestClient(
            brew_view.config.web.public_fqdn,
            brew_view.config.web.port,
            ssl_enabled=brew_view.config.web.ssl.enabled,
            url_prefix=brew_view.config.web.url_prefix,
        )

    def _event_prepare(self, event, **kwargs):

        if event.name.startswith("REQUEST"):
            request = kwargs.pop("request", None)
            if request:
                event.metadata["entity_url"] = self._client.request_url + str(
                    request.id
                )
                event.payload = {
                    k: str(getattr(request, k))
                    for k in [
                        "id",
                        "command",
                        "system",
                        "system_version",
                        "instance_name",
                    ]
                }

        elif event.name.startswith("SYSTEM"):
            system = kwargs.pop("system", None)
            if system:
                event.metadata["entity_url"] = self._client.system_url + str(system.id)
                event.payload = {"id": str(system.id)}

        elif event.name.startswith("INSTANCE"):
            instance = kwargs.pop("instance", None)
            if instance:
                event.metadata["entity_url"] = self._client.instance_url + str(
                    instance.id
                )
                event.payload = {"id": str(instance.id)}

        return event


class HttpPublisher(EventPublisher):
    def __init__(self, urls, ssl_context=None):
        self._client = AsyncHTTPClient(
            defaults={"ssl_options": ssl_context} if ssl_context else None
        )
        self._urls = urls

    def publish(self, message, **kwargs):
        for url in self._urls:
            self._client.fetch(
                url,
                raise_error=False,
                method="POST",
                headers={"content-type": "application/json"},
                body=message,
            )


class RequestPublisher(EventPublisher):
    def __init__(self, ssl_context=None):
        self._client = AsyncHTTPClient(
            defaults={"ssl_options": ssl_context} if ssl_context else None
        )

    def publish(self, message, **kwargs):
        for url in kwargs.get("urls", []):
            self._client.fetch(
                url,
                raise_error=False,
                method="POST",
                headers={"content-type": "application/json"},
                body=message,
            )

    def _event_publish_args(self, event, **kwargs):
        urls = []

        if event.name.startswith("REQUEST") and "request" in kwargs:
            if "webhooks" in kwargs["request"].metadata:
                webhook_dict = kwargs["request"].metadata["webhooks"]

                # Always add current event urls
                urls += webhook_dict.get(event.name, [])

                # Additionally do some quick translation for success / failure
                # completion events
                if event.name == Events.REQUEST_COMPLETED.name:
                    urls += webhook_dict.get(
                        "REQUEST_FAILURE" if event.error else "REQUEST_SUCCESS", []
                    )

        return {"urls": urls} if urls else {}


class WebsocketPublisher(BeergardenPublisher):
    """Publisher implementation that publishes to a websocket"""

    def __init__(self, socket_class):
        BeergardenPublisher.__init__(self)

        self._socket = socket_class

    def publish(self, message, **kwargs):
        self._socket.publish(message)

    def shutdown(self):
        self._socket.shutdown()


class MongoPublisher(BeergardenPublisher):
    """Publisher implementation that 'publishes' to Mongo"""

    def publish(self, message, **kwargs):
        message.save()

    def _event_serialize(self, event, **kwargs):
        return MongoParser.parse_event(
            MongoParser.serialize_event(event, to_string=False)
        )


class TornadoPikaPublisher(BeergardenPublisher, PikaClient):
    def __init__(self, **kwargs):
        self.logger = logging.getLogger(__name__)

        self._shutdown_timeout = timedelta(seconds=kwargs.pop("shutdown_timeout", 5))
        self._work_queue = Queue()
        self._connection = None
        self._channel = None

        self.coroutiner = CoroutineMaker(
            {"TornadoConnection": "on_open_callback", "channel": "on_open_callback"}
        )

        # Trying to get super() to work with incompatible signatures is a nightmare
        BeergardenPublisher.__init__(self)
        PikaClient.__init__(self, **kwargs)

        IOLoop.current().spawn_callback(self._process)

    def shutdown(self):
        return self._work_queue.join(timeout=self._shutdown_timeout)

    @coroutine
    def _open_connection(self):
        self._connection = yield self.coroutiner.convert(TornadoConnection)(
            parameters=self._conn_params, stop_ioloop_on_close=False
        )

    @coroutine
    def _open_channel(self):
        self._channel = yield self.coroutiner.convert(self._connection.channel)()

    @coroutine
    def _process(self):

        while True:
            item = yield self._work_queue.get()

            try:
                if not self._connection or not self._connection.is_open:
                    yield self._open_connection()
                if not self._channel or not self._channel.is_open:
                    yield self._open_channel()

                yield getattr(self._channel, item[0])(**item[1])
            finally:
                self._work_queue.task_done()

    def publish(self, message, **kwargs):
        """Publish a message.

        :param message: The message to publish
        :param kwargs: Additional message properties
        :Keyword Arguments:
            * *routing_key* --
              Routing key to use when publishing
            * *headers* --
              Headers to be included as part of the message properties
            * *expiration* --
              Expiration to be included as part of the message properties
        :return: None
        """
        self._work_queue.put(
            (
                "basic_publish",
                {
                    "exchange": self._exchange,
                    "routing_key": kwargs["routing_key"],
                    "body": message,
                    "properties": BasicProperties(
                        app_id="beer-garden",
                        content_type="text/plain",
                        headers=kwargs.pop("headers", None),
                        expiration=kwargs.pop("expiration", None),
                    ),
                },
            )
        )

    def _event_publish_args(self, event, **kwargs):

        # Main thing we need to do here is figure out the appropriate routing key
        args = {}
        if event.metadata and "routing_key" in event.metadata:
            args["routing_key"] = event.metadata["routing_key"]
        elif "request" in kwargs:
            request = kwargs["request"]
            args["routing_key"] = get_routing_key(
                "request", request.system, request.system_version, request.instance_name
            )
        else:
            args["routing_key"] = "beergarden"

        return args


class CoroutineMaker(object):
    """Helper class to wrap functions in Tornado futures

    Tornado has a nice @return_future decorator that converts a callback-based function
    into a Tornado async one. The issue is that it expects the wrapped function to take
    a `callback` argument and invoke that argument when execution is completed.

    We need to convert a couple of pika functions in this way. Unfortunately, they're
    expecting a callback function to be provided with the `on_open_callback` keyword
    parameter. So we use this class to wrap the functions and mangle the arguments when
    the function is called to pass the Tornado-provided callback to the function like
    it's expecting.
    """

    def __init__(self, config):
        self._signatures = {}

        for func, param in config.items():
            if isinstance(param, int):
                self._signatures[func] = self._positional_signature(param)

            elif isinstance(param, string_types):
                self._signatures[func] = self._keyword_signature(param)

            else:
                raise Exception("Bad configuration")

    def convert(self, func):
        @return_future
        @wraps(func)
        def wrapper(*args, **kwargs):
            try:
                args, kwargs = self._signatures[func.__name__](args, kwargs)
            except KeyError:
                pass
            func(*args, **kwargs)

        return wrapper

    @staticmethod
    def _positional_signature(position):
        def signature(args, kwargs):
            args = list(args)
            args.insert(position, kwargs.pop("callback"))
            return args, kwargs

        return signature

    @staticmethod
    def _keyword_signature(keyword):
        def signature(args, kwargs):
            kwargs[keyword] = kwargs.pop("callback")
            return args, kwargs

        return signature



import logging

import mongoengine.errors

from bg_utils.mongo.models import Command
from bg_utils.mongo.parser import MongoParser
from brew_view.authorization import authenticated, Permissions
from brew_view.base_handler import BaseHandler


class CommandListAPI(BaseHandler):

    parser = MongoParser()
    logger = logging.getLogger(__name__)

    @authenticated(permissions=[Permissions.COMMAND_READ])
    def get(self):
        """
        ---
        summary: Retrieve all Commands
        responses:
          200:
            description: All Commands
            schema:
              type: array
              items:
                $ref: '#/definitions/Command'
          50x:
            $ref: '#/definitions/50xError'
        tags:
          - Commands
        """
        self.logger.debug("Getting Commands")

        self.set_header("Content-Type", "application/json; charset=UTF-8")
        try:
            self.write(
                self.parser.serialize_command(
                    Command.objects.all(), many=True, to_string=True
                )
            )
        except mongoengine.errors.DoesNotExist as ex:
            self.logger.error(
                "Got an error while attempting to serialize commands. "
                "This error usually indicates "
                "there are orphans in the database."
            )
            raise mongoengine.errors.InvalidDocumentError(ex)



import json
import logging
from datetime import timedelta
from functools import reduce

from mongoengine import Q
from tornado.gen import coroutine
from tornado.locks import Event
from tornado.util import TimeoutError

import bg_utils
import brew_view
from bg_utils.mongo.models import Request, System
from bg_utils.mongo.parser import MongoParser
from brew_view import thrift_context
from brew_view.authorization import authenticated, Permissions
from brew_view.base_handler import BaseHandler
from brew_view.metrics import request_created, http_api_latency_total, request_latency
from brewtils.errors import (
    ModelValidationError,
    RequestPublishException,
    TimeoutExceededError,
)
from brewtils.models import Events


class RequestListAPI(BaseHandler):

    parser = MongoParser()
    logger = logging.getLogger(__name__)

    indexes = [index["name"] for index in Request._meta["indexes"]]

    @authenticated(permissions=[Permissions.REQUEST_READ])
    def get(self):
        """
        ---
        summary: Retrieve a page of all Requests
        description: |
          This endpoint queries multiple requests at once. Because it's intended to be
          used with Datatables the query parameters are ... complicated. Here are
          things to keep in mind:

          * With no query parameters this endpoint will return the first 100 non-child
            requests. This can be controlled by passing the `start` and `length` query
            parameters.

          * This endpoint does NOT return child request definitions. If you want to see
            child requests you must use the /api/v1/requests/{request_id} endpoint.

          * By default this endpoint also does not include child requests in the
            response. That is, if a request has a non-null `parent` field it will not
            be included in the response array. Use the `include_children` query
            parameter to change this behavior.

          To filter, search, and order you need to conform to how Datatables structures
          its query parameters.

          * To indicate fields that should be included in the response specify multiple
          `columns` query parameters:
          ```JSON
          {
            "data": "command",
            "name": "",
            "searchable": true,
            "orderable": true,
            "search": {"value":"","regex":false}
          }
          {
            "data": "system",
            "name": "",
            "searchable": true,
            "orderable": true,
            "search": {"value": "","regex": false}
          }
          ```
          * To filter a specific field set the value in the `search` key of its
            `column` definition:
          ```JSON
          {
            "data": "status",
            "name": "",
            "searchable": true,
            "orderable": true,
            "search": {"value": "SUCCESS", "regex":false}
          }
          ```

          * To sort by a field use the `order` parameter. The `column` value should be
            the index of the column to sort and the `dir` value can be either "asc" or
            "desc."
          ```JSON
          {"column": 3,"dir": "asc"}
          ```

          * To perform a text-based search across all fields use the `search` parameter:
          ```JSON
          { "value": "SEARCH VALUE", "regex": false }
          ```
        parameters:
          - name: include_children
            in: query
            required: false
            description: |
                Flag indicating whether to include child requests in the response list
            type: boolean
            default: false
          - name: start
            in: query
            required: false
            description: The starting index for the page
            type: integer
          - name: length
            in: query
            required: false
            description: The maximum number of Requests to include in the page
            type: integer
          - name: draw
            in: query
            required: false
            description: Used by datatables, will be echoed in a response header
            type: integer
          - name: columns
            in: query
            required: false
            description: Datatables column definitions
            type: array
            collectionFormat: multi
            items:
              properties:
                data:
                  type: string
                name:
                  type: string
                searchable:
                  type: boolean
                  default: true
                orderable:
                  type: boolean
                  default: true
                search:
                  properties:
                    value:
                      type: string
                    regex:
                      type: boolean
                      default: false
          - name: search
            in: query
            required: false
            description: Datatables search object
            type: string
          - name: order
            in: query
            required: false
            description: Datatables order object
            type: string
        responses:
          200:
            description: A page of Requests
            schema:
              type: array
              items:
                $ref: '#/definitions/Request'
            headers:
              start:
                type: integer
                description: Echo of 'start' query parameter or '0'
              length:
                type: integer
                description: Number of Requests in the response
              draw:
                type: integer
                description: Echo of the 'draw' query parameter
              recordsFiltered:
                type: integer
                description: The number of Requests that satisfied the search filters
              recordsTotal:
                type: integer
                description: The total number of Requests
          50x:
            $ref: '#/definitions/50xError'
        tags:
          - Requests
        """
        self.logger.debug("Getting Requests")

        query_set, requested_fields = self._get_query_set()

        # Actually execute the query. The slicing greatly reduces load time.
        start = int(self.get_argument("start", default=0))
        length = int(self.get_argument("length", default=100))
        requests = query_set[start : start + length]

        # Sweet, we have data. Now setup some headers for the response
        response_headers = {
            # These are a courtesy for non-datatables requests. We want people
            # making a request with no headers to realize they probably aren't
            # getting the full dataset
            "start": start,
            "length": len(requests),
            # And these are required by datatables
            "recordsFiltered": query_set.count(),  # This is another query
            "recordsTotal": Request.objects.count(),
            "draw": self.get_argument("draw", ""),
        }

        for key, value in response_headers.items():
            self.add_header(key, value)
            self.add_header("Access-Control-Expose-Headers", key)

        self.set_header("Content-Type", "application/json; charset=UTF-8")
        self.write(
            self.parser.serialize_request(
                requests, to_string=True, many=True, only=requested_fields
            )
        )

    @coroutine
    @authenticated(permissions=[Permissions.REQUEST_CREATE])
    def post(self):
        """
        ---
        summary: Create a new Request
        parameters:
          - name: request
            in: body
            description: The Request definition
            schema:
              $ref: '#/definitions/Request'
          - name: blocking
            in: query
            required: false
            description: Flag indicating whether to wait for request completion
            type: boolean
            default: false
          - name: timeout
            in: query
            required: false
            description: Maximum time (seconds) to wait for request completion
            type: integer
            default: None (Wait forever)
        consumes:
          - application/json
          - application/x-www-form-urlencoded
        responses:
          201:
            description: A new Request has been created
            schema:
              $ref: '#/definitions/Request'
            headers:
              Instance-Status:
                type: string
                description: |
                    Current status of the Instance that will process the
                    created Request
          400:
            $ref: '#/definitions/400Error'
          50x:
            $ref: '#/definitions/50xError'
        tags:
          - Requests
        """
        self.request.event.name = Events.REQUEST_CREATED.name

        if self.request.mime_type == "application/json":
            request_model = self.parser.parse_request(
                self.request.decoded_body, from_string=True
            )
        elif self.request.mime_type == "application/x-www-form-urlencoded":
            args = {"parameters": {}}
            for key, value in self.request.body_arguments.items():
                if key.startswith("parameters."):
                    args["parameters"][key.replace("parameters.", "")] = value[
                        0
                    ].decode(self.request.charset)
                else:
                    args[key] = value[0].decode(self.request.charset)
            request_model = Request(**args)
        else:
            raise ModelValidationError("Unsupported or missing content-type header")

        if request_model.parent:
            request_model.parent = Request.objects.get(id=str(request_model.parent.id))
            request_model.has_parent = True
        else:
            request_model.has_parent = False

        if self.current_user:
            request_model.requester = self.current_user.username

        # Ok, ready to save
        request_model.save()
        request_id = str(request_model.id)

        # Set up the wait event BEFORE yielding the processRequest call
        blocking = self.get_argument("blocking", default="").lower() == "true"
        if blocking:
            brew_view.request_map[request_id] = Event()

        with thrift_context() as client:
            try:
                yield client.processRequest(request_id)
            except bg_utils.bg_thrift.InvalidRequest as ex:
                request_model.delete()
                raise ModelValidationError(ex.message)
            except bg_utils.bg_thrift.PublishException as ex:
                request_model.delete()
                raise RequestPublishException(ex.message)
            except Exception:
                if request_model.id:
                    request_model.delete()
                raise

        # Query for request from body id
        req = Request.objects.get(id=request_id)

        # Now attempt to add the instance status as a header.
        # The Request is already created at this point so it's a best-effort thing
        self.set_header("Instance-Status", "UNKNOWN")

        try:
            # Since request has system info we can query for a system object
            system = System.objects.get(name=req.system, version=req.system_version)

            # Loop through all instances in the system until we find the instance that
            # matches the request instance
            for instance in system.instances:
                if instance.name == req.instance_name:
                    self.set_header("Instance-Status", instance.status)

        # The Request is already created at this point so adding the Instance status
        # header is a best-effort thing
        except Exception as ex:
            self.logger.exception(
                "Unable to get Instance status for Request %s: %s", request_id, ex
            )

        self.request.event_extras = {"request": req}

        # Metrics
        request_created(request_model)

        if blocking:
            # Publish metrics and event here here so they aren't skewed
            # See https://github.com/beer-garden/beer-garden/issues/190
            self.request.publish_metrics = False
            http_api_latency_total.labels(
                method=self.request.method.upper(),
                route=self.prometheus_endpoint,
                status=self.get_status(),
            ).observe(request_latency(self.request.created_time))

            self.request.publish_event = False
            brew_view.event_publishers.publish_event(
                self.request.event, **self.request.event_extras
            )

            try:
                timeout = self.get_argument("timeout", default=None)
                delta = timedelta(seconds=int(timeout)) if timeout else None

                event = brew_view.request_map.get(request_id)

                yield event.wait(delta)

                request_model.reload()
            except TimeoutError:
                raise TimeoutExceededError()
            finally:
                brew_view.request_map.pop(request_id, None)

        self.set_status(201)
        self.write(self.parser.serialize_request(request_model, to_string=False))

    def _get_query_set(self):
        """Get Requests matching the HTTP request query parameters.

        :return query_set: The QuerySet representing this query
        :return requested_fields: The fields to be returned for each Request
        """
        search_params = []
        requested_fields = []
        order_by = None
        overall_search = None
        hint = []

        query_set = Request.objects

        raw_columns = self.get_query_arguments("columns")
        if raw_columns:
            columns = []

            for raw_column in raw_columns:
                column = json.loads(raw_column)
                columns.append(column)

                if column["data"]:
                    requested_fields.append(column["data"])

                if (
                    "searchable" in column
                    and column["searchable"]
                    and column["search"]["value"]
                ):
                    if column["data"] in ["created_at", "updated_at"]:
                        search_dates = column["search"]["value"].split("~")
                        start_query = Q()
                        end_query = Q()

                        if search_dates[0]:
                            start_query = Q(
                                **{column["data"] + "__gte": search_dates[0]}
                            )
                        if search_dates[1]:
                            end_query = Q(**{column["data"] + "__lte": search_dates[1]})

                        search_query = start_query & end_query
                    elif column["data"] == "status":
                        search_query = Q(
                            **{column["data"] + "__exact": column["search"]["value"]}
                        )
                    elif column["data"] == "comment":
                        search_query = Q(
                            **{column["data"] + "__contains": column["search"]["value"]}
                        )
                    else:
                        search_query = Q(
                            **{
                                column["data"]
                                + "__startswith": column["search"]["value"]
                            }
                        )

                    search_params.append(search_query)
                    hint.append(column["data"])

            raw_order = self.get_query_argument("order", default=None)
            if raw_order:
                order = json.loads(raw_order)
                order_by = columns[order.get("column")]["data"]

                hint.append(order_by)

                if order.get("dir") == "desc":
                    order_by = "-" + order_by

        raw_search = self.get_query_argument("search", default=None)
        if raw_search:
            search = json.loads(raw_search)
            if search["value"]:
                overall_search = '"' + search["value"] + '"'

        # Default to only top-level requests
        if (
            self.get_query_argument("include_children", default="false").lower()
            != "true"
        ):
            search_params.append(Q(has_parent=False))

        # Now we can construct the actual query parameters
        query_params = reduce(lambda x, y: x & y, search_params, Q())
        query_set = query_set.filter(query_params)

        # Further modify the QuerySet
        if overall_search:
            query_set = query_set.search_text(overall_search)

        if order_by:
            query_set = query_set.order_by(order_by)

        # Marshmallow treats [] as 'serialize nothing' which is not what we
        # want, so translate to None
        if requested_fields:
            query_set = query_set.only(*requested_fields)
        else:
            requested_fields = None

        # Mongo seems to prefer using only the ['parent', '<sort field>']
        # index, even when also filtering. So we have to help it a bit.
        real_hint = ["parent"]
        if "created_at" in hint:
            real_hint.append("created_at")
        for index in ["command", "system", "instance_name", "status"]:
            if index in hint:
                real_hint.append(index)
                break
        real_hint.append("index")

        # Sanity check - if index is 'bad' just let mongo deal with it
        index_name = "_".join(real_hint)
        if index_name in self.indexes:
            query_set = query_set.hint(index_name)

        return query_set, requested_fields



import re

from setuptools import setup, find_packages


def find_version(version_file):
    version_line = open(version_file, "rt").read()
    match_object = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_line, re.M)

    if not match_object:
        raise RuntimeError("Unable to find version string in %s" % version_file)

    return match_object.group(1)


with open("README.rst") as readme_file:
    readme = readme_file.read()


setup(
    name="brew-view",
    version=find_version("brew_view/_version.py"),
    description="Beergarden Frontend",
    long_description=readme,
    url="https://beer-garden.io/",
    author="The beer-garden Team",
    author_email=" ",
    license="MIT",
    packages=(find_packages(exclude=["test", "test.*"])),
    include_package_data=True,
    install_requires=[
        "apispec==0.38.0",
        "apscheduler==3.5.1",
        "bg-utils>=2.4.0",
        "brewtils>=2.4.0",
        "prometheus_client==0.3.1",
        "tornado==5.1",
    ],
    extras_require={':python_version=="2.7"': ["futures>=3.1.1"]},
    entry_points={
        "console_scripts": [
            "generate_brew_view_config=brew_view.__main__:generate_config",
            "migrate_brew_view_config=brew_view.__main__:migrate_config",
            "generate_brew_view_log_config=brew_view.__main__:generate_logging_config",
            "brew-view=brew_view.__main__:main",
        ]
    },
)



import logging
import brew_view
from bg_utils.mongo.parser import MongoParser
from brew_view.base_handler import BaseHandler
from brewtils.errors import ModelValidationError

from tornado.gen import coroutine


class LoggingConfigAPI(BaseHandler):

    parser = MongoParser()
    logger = logging.getLogger(__name__)

    def get(self):
        """
        ---
        summary: Get the plugin logging configuration
        parameters:
          - name: system_name
            in: query
            required: false
            description: Specific system name to get logging configuration
            type: string
        responses:
          200:
            description: Logging Configuration for system
            schema:
                $ref: '#/definitions/LoggingConfig'
          50x:
            $ref: '#/definitions/50xError'
        tags:
          - Config
        """
        system_name = self.get_query_argument("system_name", default=None)
        log_config = brew_view.plugin_logging_config.get_plugin_log_config(
            system_name=system_name
        )
        self.write(self.parser.serialize_logging_config(log_config, to_string=False))

    @coroutine
    def patch(self):
        """
        ---
        summary: Reload the plugin logging configuration
        description: |
          The body of the request needs to contain a set of instructions detailing the
          operation to make. Currently supported operations are below:
          ```JSON
          {
            "operations": [
                { "operation": "reload" }
            ]
          }
          ```
        parameters:
          - name: patch
            in: body
            required: true
            description: Operation to perform
            schema:
              $ref: '#/definitions/Patch'
        responses:
          200:
            description: Updated plugin logging configuration
            schema:
              $ref: '#/definitions/LoggingConfig'
          50x:
            $ref: '#/definitions/50xError'
        tags:
          - Config
        """
        operations = self.parser.parse_patch(
            self.request.decoded_body, many=True, from_string=True
        )

        for op in operations:
            if op.operation == "reload":
                brew_view.load_plugin_logging_config(brew_view.config)
            else:
                error_msg = "Unsupported operation '%s'" % op.operation
                self.logger.warning(error_msg)
                raise ModelValidationError("value", error_msg)

        self.set_status(200)
        self.write(
            self.parser.serialize_logging_config(
                brew_view.plugin_logging_config, to_string=False
            )
        )



# -*- coding: utf-8 -*-
"""module for controller for /jobs endpoint."""
import logging

from tornado.gen import coroutine

import brew_view
from bg_utils.mongo.models import Job
from bg_utils.mongo.parser import MongoParser
from brew_view.authorization import authenticated, Permissions
from brew_view.base_handler import BaseHandler
from brew_view.scheduler.runner import run_job
from brewtils.schemas import JobSchema


class JobListAPI(BaseHandler):

    parser = MongoParser()
    logger = logging.getLogger(__name__)

    @authenticated(permissions=[Permissions.JOB_READ])
    def get(self):
        """
        ---
        summary: Retrieve all Jobs.
        responses:
          200:
            description: Successfully retrieved all systems.
            schema:
              type: array
              items:
                $ref: '#/definitions/Job'
          50x:
            $ref: '#/definitions/50xError'
        tags:
          - Jobs
        """
        filter_params = {}
        for key in self.request.arguments.keys():
            if key in JobSchema.get_attribute_names():
                filter_params[key] = self.get_query_argument(key)

        self.set_header("Content-Type", "application/json; charset=UTF-8")
        self.write(
            self.parser.serialize_job(
                Job.objects.filter(**filter_params), to_string=True, many=True
            )
        )

    @coroutine
    @authenticated(permissions=[Permissions.JOB_CREATE])
    def post(self):
        """
        ---
        summary: Schedules a Job to be run.
        description: |
          Given a job, it will be scheduled to run on the interval
          set in the trigger argument.
        parameters:
          - name: job
            in: body
            description: The Job to create/schedule
            schema:
              $ref: '#/definitions/Job'
        responses:
          201:
            description: A new job has been created
            schema:
              $ref: '#/definitions/Job'
          400:
            $ref: '#/definitions/400Error'
          50x:
            $ref: '#/definitions/50xError'
        tags:
          - Jobs
        """
        document = self.parser.parse_job(self.request.decoded_body, from_string=True)
        # We have to save here, because we need an ID to pass
        # to the scheduler.
        document.save()

        try:
            brew_view.request_scheduler.add_job(
                run_job,
                None,
                kwargs={
                    "request_template": document.request_template,
                    "job_id": str(document.id),
                },
                name=document.name,
                misfire_grace_time=document.misfire_grace_time,
                coalesce=document.coalesce,
                max_instances=3,
                jobstore="beer_garden",
                replace_existing=False,
                id=str(document.id),
            )
        except Exception:
            document.delete()
            raise

        self.set_status(201)
        self.set_header("Content-Type", "application/json; charset=UTF-8")
        self.write(self.parser.serialize_job(document, to_string=False))



import logging

from tornado.gen import coroutine

from bg_utils.mongo.models import System
from bg_utils.mongo.parser import MongoParser
from brew_view import thrift_context
from brew_view.authorization import authenticated, Permissions
from brew_view.base_handler import BaseHandler
from brewtils.errors import ModelValidationError
from brewtils.models import Events


class SystemAPI(BaseHandler):

    parser = MongoParser()
    logger = logging.getLogger(__name__)

    @authenticated(permissions=[Permissions.SYSTEM_READ])
    def get(self, system_id):
        """
        ---
        summary: Retrieve a specific System
        parameters:
          - name: system_id
            in: path
            required: true
            description: The ID of the System
            type: string
          - name: include_commands
            in: query
            required: false
            description: Include the System's commands in the response
            type: boolean
            default: true
        responses:
          200:
            description: System with the given ID
            schema:
              $ref: '#/definitions/System'
          404:
            $ref: '#/definitions/404Error'
          50x:
            $ref: '#/definitions/50xError'
        tags:
          - Systems
        """
        self.logger.debug("Getting System: %s", system_id)

        include_commands = (
            self.get_query_argument("include_commands", default="true").lower()
            != "false"
        )
        self.write(
            self.parser.serialize_system(
                System.objects.get(id=system_id),
                to_string=False,
                include_commands=include_commands,
            )
        )

    @coroutine
    @authenticated(permissions=[Permissions.SYSTEM_DELETE])
    def delete(self, system_id):
        """
        Will give Bartender a chance to remove instances of this system from the
        registry but will always delete the system regardless of whether the Bartender
        operation succeeds.
        ---
        summary: Delete a specific System
        description: Will remove instances of local plugins from the registry, clear
            and remove message queues, and remove the system from the database.
        parameters:
          - name: system_id
            in: path
            required: true
            description: The ID of the System
            type: string
        responses:
          204:
            description: System has been successfully deleted
          404:
            $ref: '#/definitions/404Error'
          50x:
            $ref: '#/definitions/50xError'
        tags:
          - Systems
        """
        self.request.event.name = Events.SYSTEM_REMOVED.name
        self.request.event_extras = {"system": System.objects.get(id=system_id)}

        with thrift_context() as client:
            yield client.removeSystem(str(system_id))

        self.set_status(204)

    @coroutine
    @authenticated(permissions=[Permissions.SYSTEM_UPDATE])
    def patch(self, system_id):
        """
        ---
        summary: Partially update a System
        description: |
          The body of the request needs to contain a set of instructions detailing the
          updates to apply.
          Currently supported operations are below:
          ```JSON
          {
            "operations": [
              { "operation": "replace", "path": "/commands", "value": "" },
              { "operation": "replace", "path": "/description", "value": "new description"},
              { "operation": "replace", "path": "/display_name", "value": "new display name"},
              { "operation": "replace", "path": "/icon_name", "value": "new icon name"},
              { "operation": "update", "path": "/metadata", "value": {"foo": "bar"}}
            ]
          }
          ```
          Where `value` is a list of new Commands.
        parameters:
          - name: system_id
            in: path
            required: true
            description: The ID of the System
            type: string
          - name: patch
            in: body
            required: true
            description: Instructions for how to update the System
            schema:
              $ref: '#/definitions/Patch'
        responses:
          200:
            description: System with the given ID
            schema:
              $ref: '#/definitions/System'
          400:
            $ref: '#/definitions/400Error'
          404:
            $ref: '#/definitions/404Error'
          50x:
            $ref: '#/definitions/50xError'
        tags:
          - Systems
        """
        self.request.event.name = Events.SYSTEM_UPDATED.name

        system = System.objects.get(id=system_id)
        operations = self.parser.parse_patch(
            self.request.decoded_body, many=True, from_string=True
        )

        for op in operations:
            if op.operation == "replace":
                if op.path == "/commands":
                    new_commands = self.parser.parse_command(op.value, many=True)

                    if (
                        system.commands
                        and "dev" not in system.version
                        and system.has_different_commands(new_commands)
                    ):
                        raise ModelValidationError(
                            "System %s-%s already exists with different commands"
                            % (system.name, system.version)
                        )

                    system.upsert_commands(new_commands)
                elif op.path in ["/description", "/icon_name", "/display_name"]:
                    if op.value is None:
                        # If we set an attribute to None, mongoengine marks that
                        # attribute for deletion, so we don't do that.
                        value = ""
                    else:
                        value = op.value
                    attr = op.path.strip("/")
                    self.logger.debug("Updating system %s" % attr)
                    self.logger.debug("Old: %s" % getattr(system, attr))
                    setattr(system, attr, value)
                    self.logger.debug("Updated: %s" % getattr(system, attr))
                    system.save()
                else:
                    error_msg = "Unsupported path '%s'" % op.path
                    self.logger.warning(error_msg)
                    raise ModelValidationError("value", error_msg)
            elif op.operation == "update":
                if op.path == "/metadata":
                    self.logger.debug("Updating system metadata")
                    self.logger.debug("Old: %s" % system.metadata)
                    system.metadata.update(op.value)
                    self.logger.debug("Updated: %s" % system.metadata)
                    system.save()
                else:
                    error_msg = "Unsupported path for update '%s'" % op.path
                    self.logger.warning(error_msg)
                    raise ModelValidationError("path", error_msg)
            elif op.operation == "reload":
                with thrift_context() as client:
                    yield client.reloadSystem(system_id)
            else:
                error_msg = "Unsupported operation '%s'" % op.operation
                self.logger.warning(error_msg)
                raise ModelValidationError("value", error_msg)

        system.reload()

        self.request.event_extras = {"system": system, "patch": operations}

        self.write(self.parser.serialize_system(system, to_string=False))
